from __future__ import annotations

import os
from functools import partial
from typing import TYPE_CHECKING, List, Literal, get_args

import yaml
from bec_lib import config_helper
from bec_lib.bec_yaml_loader import yaml_load
from bec_lib.callback_handler import EventType
from bec_lib.endpoints import MessageEndpoints
from bec_lib.file_utils import DeviceConfigWriter
from bec_lib.logger import bec_logger
from bec_lib.messages import ConfigAction, ScanStatusMessage
from bec_lib.plugin_helper import plugin_package_name, plugin_repo_path
from bec_qthemes import apply_theme, material_icon
from qtpy.QtCore import QMetaObject, Qt, QThreadPool, Signal
from qtpy.QtGui import QColor
from qtpy.QtWidgets import (
    QApplication,
    QFileDialog,
    QMessageBox,
    QPushButton,
    QTextEdit,
    QVBoxLayout,
    QWidget,
)

from bec_widgets.applications.views.device_manager_view.device_manager_dialogs import (
    ConfigChoiceDialog,
    DeviceFormDialog,
)
from bec_widgets.applications.views.device_manager_view.device_manager_dialogs.upload_redis_dialog import (
    UploadRedisDialog,
)
from bec_widgets.utils.colors import get_accent_colors
from bec_widgets.utils.error_popups import SafeSlot
from bec_widgets.utils.toolbars.actions import MaterialIconAction
from bec_widgets.utils.toolbars.bundles import ToolbarBundle
from bec_widgets.utils.toolbars.toolbar import ModularToolBar
from bec_widgets.widgets.containers.dock_area.basic_dock_area import DockAreaWidget
from bec_widgets.widgets.control.device_manager.components import (
    DeviceTable,
    DMConfigView,
    DocstringView,
    OphydValidation,
)
from bec_widgets.widgets.control.device_manager.components._util import SharedSelectionSignal
from bec_widgets.widgets.control.device_manager.components.ophyd_validation.ophyd_validation_utils import (
    ConfigStatus,
    ConnectionStatus,
)
from bec_widgets.widgets.progress.device_initialization_progress_bar.device_initialization_progress_bar import (
    DeviceInitializationProgressBar,
)
from bec_widgets.widgets.services.device_browser.device_item.config_communicator import (
    CommunicateConfigAction,
)
from bec_widgets.widgets.utility.spinner.spinner import SpinnerWidget

if TYPE_CHECKING:  # pragma: no cover
    from bec_lib.client import BECClient

logger = bec_logger.logger

_yes_no_question = partial(
    QMessageBox.question,
    buttons=QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No,
    defaultButton=QMessageBox.StandardButton.No,
)


class CustomBusyWidget(QWidget):
    """Custom busy widget to show during device config upload."""

    cancel_requested = Signal()

    def __init__(self, parent=None, client: BECClient | None = None):
        super().__init__(parent=parent)

        # Widgets
        self.progress = QWidget(parent=self)
        self.progress_layout = QVBoxLayout(self.progress)
        self.progress_layout.setContentsMargins(6, 6, 6, 6)
        self.progress_inner = DeviceInitializationProgressBar(parent=self.progress, client=client)
        self.progress_layout.addWidget(self.progress_inner)
        self.progress.setMinimumWidth(320)

        # Spinner
        self.spinner = SpinnerWidget(parent=self)
        scale = self._ui_scale()
        spinner_size = int(scale * 0.12) if scale else 1
        spinner_size = max(32, min(spinner_size, 96))
        self.spinner.setFixedSize(spinner_size, spinner_size)

        # Cancel button
        self.cancel_button = QPushButton("Cancel Upload", parent=self)
        self.cancel_button.setIcon(material_icon("cancel"))
        self.cancel_button.clicked.connect(self.cancel_requested.emit)
        button_height = int(spinner_size * 0.9)
        button_height = max(36, min(button_height, 72))
        aspect_ratio = 3.8  # width / height, visually stable for text buttons
        button_width = int(button_height * aspect_ratio)
        self.cancel_button.setFixedSize(button_width, button_height)
        color = get_accent_colors()
        self.cancel_button.setStyleSheet(f"""
            QPushButton {{
                background-color: {color.emergency.name()};
                color: white;
                font-weight: 600;
                border-radius: 6px;
            }}
            """)

        # Layout
        content_layout = QVBoxLayout(self)
        content_layout.setContentsMargins(24, 24, 24, 24)
        content_layout.setSpacing(16)
        content_layout.addStretch()
        content_layout.addWidget(self.spinner, 0, Qt.AlignmentFlag.AlignHCenter)
        content_layout.addWidget(self.progress, 0, Qt.AlignmentFlag.AlignHCenter)
        content_layout.addStretch()
        content_layout.addWidget(self.cancel_button, 0, Qt.AlignmentFlag.AlignHCenter)

        if hasattr(color, "_colors"):
            bg_color = color._colors.get("BG", None)
        if bg_color is None:  # Fallback if missing
            bg_color = QColor(50, 50, 50, 255)
        self.setStyleSheet(f"""
            background-color: {bg_color.name()};
            border-radius: 12px;
            """)

    def _ui_scale(self) -> int:
        parent = self.parent()
        if not parent:
            return 0
        return min(parent.width(), parent.height())

    def showEvent(self, event):
        """Show event to start the spinner."""
        super().showEvent(event)
        self.spinner.start()

    def hideEvent(self, event):
        """Hide event to stop the spinner."""
        super().hideEvent(event)
        self.spinner.stop()


class DeviceManagerDisplayWidget(DockAreaWidget):
    """Device Manager main display widget. This contains all sub-widgets and the toolbar."""

    RPC = False

    request_ophyd_validation = Signal(list, bool, bool)

    def __init__(self, parent=None, *args, **kwargs):
        super().__init__(parent=parent, variant="compact", *args, **kwargs)

        # State variable for config upload
        self._config_upload_active: bool = False
        self._config_in_sync: bool = False
        scan_status = self.bec_dispatcher.client.connector.get(MessageEndpoints.scan_status())
        initial_status = scan_status.status if scan_status is not None else "closed"
        self._scan_is_running: bool = initial_status in ["open", "paused"]

        # Push to Redis dialog
        self._upload_redis_dialog: UploadRedisDialog | None = None
        self._dialog_validation_connection: QMetaObject.Connection | None = None

        # NOTE: We need here a separate config helper instance to avoid conflicts with
        # other communications to REDIS as uploading a config through a CommunicationConfigAction
        # will block if we use the config_helper from self.client.config._config_helper
        self._config_helper = config_helper.ConfigHelper(self.client.connector)
        self._shared_selection = SharedSelectionSignal()

        # Device Table View widget
        self.device_table_view = DeviceTable(self)

        # Device Config View widget
        self.dm_config_view = DMConfigView(self)

        # Docstring View
        self.dm_docs_view = DocstringView(self)

        # Ophyd Test view
        self.ophyd_widget_view = QWidget(self)
        layout = QVBoxLayout(self.ophyd_widget_view)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(4)
        self.ophyd_test_view = OphydValidation(self, hide_legend=False)
        layout.addWidget(self.ophyd_test_view)

        # Validation Results view
        self.validation_results = QTextEdit(self)
        self.validation_results.setReadOnly(True)
        self.validation_results.setPlaceholderText("Validation results will appear here...")
        layout.addWidget(self.validation_results)
        self.ophyd_test_view.item_clicked.connect(self._ophyd_test_item_clicked_cb)

        for signal, slots in [
            (
                self.device_table_view.selected_devices,
                (self.dm_config_view.on_select_config, self.dm_docs_view.on_select_config),
            ),
            (
                self.ophyd_test_view.validation_completed,
                (self.device_table_view.update_device_validation,),
            ),
            (
                self.ophyd_test_view.multiple_validations_completed,
                (self.device_table_view.update_multiple_device_validations,),
            ),
            (self.request_ophyd_validation, (self.ophyd_test_view.change_device_configs,)),
            (
                self.device_table_view.device_configs_changed,
                (self.ophyd_test_view.device_table_config_changed,),
            ),
            (
                self.device_table_view.device_config_in_sync_with_redis,
                (self._update_config_in_sync,),
            ),
            (self.device_table_view.device_row_dbl_clicked, (self._edit_device_action,)),
        ]:
            for slot in slots:
                signal.connect(slot)

        self._scan_status_callback_id = self.bec_dispatcher.client.callbacks.register(
            EventType.SCAN_STATUS, self._update_scan_running
        )

        # Add toolbar
        self._add_toolbar()

        # Build dock layout using shared helpers
        self._build_docks()

    def cleanup(self):
        self.bec_dispatcher.client.callbacks.remove(self._scan_status_callback_id)
        super().cleanup()

    def closeEvent(self, event):
        """If config upload is active when application is exiting, cancel it."""
        logger.info("Application is quitting, checking for active config upload...")
        if self._config_upload_active:
            logger.info("Application is quitting, cancelling active config upload...")
            self._config_helper.send_config_request(
                action="cancel", config=None, wait_for_response=True, timeout_s=10
            )
            logger.info("Config upload cancelled.")
        super().closeEvent(event)

    ##############################
    ### Custom set busy widget ###
    ##############################

    def create_busy_state_widget(self) -> QWidget:
        """Create a custom busy state widget for uploading device configurations."""
        widget = CustomBusyWidget(parent=self, client=self.client)
        widget.cancel_requested.connect(self._cancel_device_config_upload)
        return widget

    def _set_busy_wrapper(self, enabled: bool):
        """Thin wrapper around set_busy to flip the state variable."""
        self._busy_overlay.set_opacity(0.92)
        self._config_upload_active = enabled
        self.set_busy(enabled=enabled)

    ##############################
    ### Toolbar and Dock setup ###
    ##############################

    def _add_toolbar(self):
        self.toolbar = ModularToolBar(self)

        # Add IO actions
        self._add_io_actions()
        self._add_table_actions()
        self.toolbar.show_bundles(["IO", "Table"])
        self._root_layout.insertWidget(0, self.toolbar)

    def _build_docks(self) -> None:
        # Central device table
        self.device_table_view_dock = self.new(
            self.device_table_view,
            return_dock=True,
            closable=False,
            floatable=False,
            movable=False,
            show_title_bar=False,
        )

        # Bottom area: docstrings
        self.dm_docs_view_dock = self.new(
            self.dm_docs_view,
            where="bottom",
            relative_to=self.device_table_view_dock,
            return_dock=True,
            closable=False,
            floatable=False,
            movable=False,
            show_title_bar=False,
        )
        # Config view left of docstrings
        self.dm_config_view_dock = self.new(
            self.dm_config_view,
            where="left",
            relative_to=self.dm_docs_view_dock,
            return_dock=True,
            closable=False,
            floatable=False,
            movable=False,
            show_title_bar=False,
        )

        # Right area: ophyd test + validation
        self.ophyd_test_dock_view = self.new(
            self.ophyd_widget_view,
            where="right",
            relative_to=self.device_table_view_dock,
            return_dock=True,
            closable=False,
            floatable=False,
            movable=False,
            show_title_bar=False,
        )

        self.set_layout_ratios(splitter_overrides={0: [7, 3], 1: [3, 7]})

    def _add_io_actions(self):
        # Create IO bundle
        io_bundle = ToolbarBundle("IO", self.toolbar.components)

        # Load from disk
        load = MaterialIconAction(
            text_position="under",
            icon_name="file_open",
            parent=self,
            tooltip="Load configuration file from disk",
            label_text="Load Config",
        )
        self.toolbar.components.add_safe("load", load)
        load.action.triggered.connect(self._load_file_action)
        io_bundle.add_action("load")

        # Add safe to disk
        save_to_disk = MaterialIconAction(
            text_position="under",
            icon_name="file_save",
            parent=self,
            tooltip="Save config to disk",
            label_text="Save Config",
        )
        self.toolbar.components.add_safe("save_to_disk", save_to_disk)
        save_to_disk.action.triggered.connect(self._save_to_disk_action)
        io_bundle.add_action("save_to_disk")

        # Add flush config in redis
        flush_redis = MaterialIconAction(
            text_position="under",
            icon_name="delete_sweep",
            parent=self,
            tooltip="Flush current config in BEC Server",
            label_text="Flush loaded Config",
        )
        flush_redis.action.triggered.connect(self._flush_redis_action)
        self.toolbar.components.add_safe("flush_redis", flush_redis)
        io_bundle.add_action("flush_redis")

        # Add load config from redis
        load_redis = MaterialIconAction(
            text_position="under",
            icon_name="cached",
            parent=self,
            tooltip="Load current config from BEC Server",
            label_text="Get loaded Config",
        )
        load_redis.action.triggered.connect(self._load_redis_action)
        self.toolbar.components.add_safe("load_redis", load_redis)
        io_bundle.add_action("load_redis")

        # Update config action
        update_config_redis = MaterialIconAction(
            text_position="under",
            icon_name="cloud_upload",
            parent=self,
            tooltip="Update current config in BEC Server",
            label_text="Update Config",
        )
        update_config_redis.action.setEnabled(False)

        update_config_redis.action.triggered.connect(self._update_redis_action)
        self.toolbar.components.add_safe("update_config_redis", update_config_redis)
        io_bundle.add_action("update_config_redis")

        # Add load config from plugin dir
        self.toolbar.add_bundle(io_bundle)

    # Table actions
    def _add_table_actions(self) -> None:
        table_bundle = ToolbarBundle("Table", self.toolbar.components)

        # Reset composed view
        reset_composed = MaterialIconAction(
            text_position="under",
            icon_name="delete_sweep",
            parent=self,
            tooltip="Reset current composed config view",
            label_text="Reset Config View",
        )
        reset_composed.action.triggered.connect(self._reset_composed_view)
        self.toolbar.components.add_safe("reset_composed", reset_composed)
        table_bundle.add_action("reset_composed")

        # Add device
        add_device = MaterialIconAction(
            text_position="under",
            icon_name="add",
            parent=self,
            tooltip="Add new device",
            label_text="Add Device",
        )
        add_device.action.triggered.connect(self._add_device_action)
        self.toolbar.components.add_safe("add_device", add_device)
        table_bundle.add_action("add_device")

        # Remove device
        remove_device = MaterialIconAction(
            text_position="under",
            icon_name="remove",
            parent=self,
            tooltip="Remove device",
            label_text="Remove Device",
        )
        remove_device.action.triggered.connect(self._remove_device_action)
        self.toolbar.components.add_safe("remove_device", remove_device)
        table_bundle.add_action("remove_device")

        # Rerun validation
        rerun_validation = MaterialIconAction(
            text_position="under",
            icon_name="checklist",
            parent=self,
            tooltip="Run device validation with 'connect' on selected devices",
            label_text="Validate Connection",
        )
        rerun_validation.action.triggered.connect(self._run_validate_connection)
        self.toolbar.components.add_safe("rerun_validation", rerun_validation)
        table_bundle.add_action("rerun_validation")

        # Add load config from plugin dir
        self.toolbar.add_bundle(table_bundle)

    ######################################
    ### Update button state management ###
    ######################################

    @SafeSlot(dict, dict)
    def _update_scan_running(self, scan_info: dict, _: dict):
        """disable editing when scans are running and enable editing when they are finished"""
        msg = ScanStatusMessage.model_validate(scan_info)
        self._scan_is_running = msg.status in ["open", "paused"]
        self._update_config_enabled_button()

    def _update_config_in_sync(self, in_sync: bool):
        self._config_in_sync = in_sync
        self._update_config_enabled_button()

    def _update_config_enabled_button(self):
        action = self.toolbar.components.get_action("update_config_redis")
        enabled = not self._config_in_sync and not self._scan_is_running
        action.action.setEnabled(enabled)
        if enabled:  # button is enabled
            action.action.setToolTip("Push current config to BEC Server")
        elif self._scan_is_running:
            action.action.setToolTip("Scan is currently running, config updates disabled.")
        else:
            action.action.setToolTip("Current config is in sync with BEC Server, updates disabled.")

    #######################
    ### Action Handlers ###
    #######################

    @SafeSlot()
    @SafeSlot(bool)
    def _run_validate_connection(self, connect: bool = True):
        """Action for the 'rerun_validation' action to rerun validation on selected devices."""
        configs = list(self.device_table_view.get_selected_device_configs())
        if not configs:
            configs = self.device_table_view.get_device_config()
        # Adjust the state of the icons in the device table view
        self.device_table_view.update_multiple_device_validations(
            [
                (cfg, ConfigStatus.UNKNOWN.value, ConnectionStatus.UNKNOWN.value, "")
                for cfg in configs
            ]
        )
        self.request_ophyd_validation.emit(configs, True, connect)

    @SafeSlot()
    def _load_file_action(self):
        """Action for the 'load' action to load a config from disk for the io_bundle of the toolbar."""
        config_path = self._get_config_base_path()

        # Implement the file loading logic here
        start_dir = os.path.abspath(config_path)
        file_path = self._get_file_path(start_dir, "open_file")
        if file_path:
            self._load_config_from_file(file_path)

    def _get_config_base_path(self) -> str:
        """Get the base path for device configurations."""
        try:
            plugin_path = plugin_repo_path()
            plugin_name = plugin_package_name()
            config_path = os.path.join(plugin_path, plugin_name, "device_configs")
        except ValueError:
            # Get the recovery config path as fallback
            config_path = self._get_recovery_config_path()
            logger.warning(
                f"No plugin repository installed, fallback to recovery config path: {config_path}"
            )
        return config_path

    def _get_file_path(self, start_dir: str, mode: Literal["open_file", "save_file"]) -> str:
        ALLOWED_EXTS = [".yaml", ".yml"]
        filter_str = "YAML files (*.yaml *.yml);;All Files (*)"
        initial_filter = "YAML files (*.yaml *.yml);;"
        if mode == "open_file":
            file_path, _ = QFileDialog.getOpenFileName(
                self,
                caption="Select Config File",
                dir=start_dir,
                filter=filter_str,
                selectedFilter=initial_filter,
            )
        else:
            file_path, _ = QFileDialog.getSaveFileName(
                self,
                caption="Save Config File",
                dir=start_dir,
                filter=filter_str,
                selectedFilter=initial_filter,
            )
        if not file_path:
            return ""
        _, ext = os.path.splitext(file_path)
        if ext.lower() not in ALLOWED_EXTS:
            file_path += ".yaml"
        return file_path

    def _load_config_from_file(self, file_path: str):
        """
        Load device config from a given file path and update the device table view.

        Args:
            file_path (str): Path to the configuration file.
        """
        try:
            config = [{"name": k, **v} for k, v in yaml_load(file_path).items()]
        except Exception as e:
            logger.error(f"Failed to load config from file {file_path}. Error: {e}")
            return
        self._open_config_choice_dialog(config)

    def _open_config_choice_dialog(self, config: List[dict]):
        """
        Open a dialog to choose whether to replace or add the loaded config.

        Args:
            config (List[dict]): List of device configurations loaded from the file.
        """
        if len(self.device_table_view.get_device_config()) == 0:
            # If no config is composed yet, load directly
            self.device_table_view.set_device_config(config)
            return
        dialog = ConfigChoiceDialog(self)
        result = dialog.exec()
        if result == ConfigChoiceDialog.Result.REPLACE:
            self.device_table_view.set_device_config(config)
        elif result == ConfigChoiceDialog.Result.ADD:
            self.device_table_view.add_device_configs(config)

    @SafeSlot()
    def _flush_redis_action(self):
        """Action to flush the current config in Redis."""
        if self.client.device_manager is None:
            logger.error("No device manager connected, cannot load config from BEC Server.")
            return
        if len(self.client.device_manager.devices) == 0:
            logger.info("No devices in BEC Server, nothing to flush.")
            QMessageBox.information(
                self, "No Devices", "There is currently no config loaded on the BEC Server."
            )
            return
        reply = _yes_no_question(
            self,
            "Flush BEC Server Config",
            "Do you really want to flush the current config in BEC Server?",
        )
        if reply == QMessageBox.StandardButton.Yes:
            self.client.config.reset_config()
            logger.info("Successfully flushed configuration in BEC Server.")
            # Check if config is in sync, enable load redis button
            self.device_table_view.device_config_in_sync_with_redis.emit(
                self.device_table_view._is_config_in_sync_with_redis()
            )
            validation_results = self.device_table_view.get_validation_results()
            for config, config_status, connection_status in validation_results.values():
                if connection_status == ConnectionStatus.CONNECTED.value:
                    self.device_table_view.update_device_validation(
                        config, config_status, ConnectionStatus.CAN_CONNECT, ""
                    )

    @SafeSlot()
    def _load_redis_action(self):
        """Action for the 'load_redis' action to load the current config from Redis for the io_bundle of the toolbar."""
        if self.client.device_manager is None:
            logger.error("No device manager connected, cannot load config from BEC Server.")
            return
        if not self.device_table_view.get_device_config():
            # If no config is composed yet, load directly
            self.device_table_view.set_device_config(
                self.client.device_manager._get_redis_device_config()
            )
            return
        reply = _yes_no_question(
            self,
            "Load currently active config in BEC Server",
            "Do you really want to discard the current config and reload?",
        )
        if reply == QMessageBox.StandardButton.Yes:
            self.device_table_view.set_device_config(
                self.client.device_manager._get_redis_device_config()
            )

    @SafeSlot()
    def _update_redis_action(self) -> None | QMessageBox.StandardButton:
        """Action to push the current composition to Redis using the upload dialog."""
        # Check if validations are still running
        if self.ophyd_test_view.running_ophyd_tests is True:
            return QMessageBox.warning(
                self, "Validation in Progress", "Please wait for the validation to finish."
            )

        # Get all device configurations with their validation status
        validation_results = self.device_table_view.get_validation_results()
        # Create and show upload dialog
        self._upload_redis_dialog = UploadRedisDialog(
            parent=self, device_configs=validation_results
        )
        self._upload_redis_dialog.request_ophyd_validation.connect(
            self.request_ophyd_validation.emit
        )

        # Show dialog
        reply = self._upload_redis_dialog.exec_()

        if reply == UploadRedisDialog.UploadAction.OK:
            self._push_composition_to_redis(action="set")
        elif reply == UploadRedisDialog.UploadAction.CANCEL:
            self.ophyd_test_view.cancel_all_validations()
        elif reply == UploadRedisDialog.UploadAction.CONNECTION_TEST_REQUESTED:
            return QMessageBox.information(
                self, "Connection Test Requested", "Running connection test on untested devices."
            )

    def _push_composition_to_redis(self, action: ConfigAction):
        """Push the current device composition to Redis."""
        if action not in get_args(ConfigAction):
            logger.error(f"Invalid config action: {action} for uploading to BEC Server.")
            return
        config = {cfg.pop("name"): cfg for cfg in self.device_table_view.get_device_config()}
        threadpool = QThreadPool.globalInstance()
        comm = CommunicateConfigAction(self._config_helper, None, config, action)
        comm.signals.done.connect(self._handle_push_complete_to_communicator)
        comm.signals.error.connect(self._handle_exception_from_communicator)
        threadpool.start(comm)
        self._set_busy_wrapper(enabled=True)

    def _cancel_device_config_upload(self):
        """Cancel the device configuration upload process."""
        threadpool = QThreadPool.globalInstance()
        comm = CommunicateConfigAction(self._config_helper, None, {}, "cancel")
        # Cancelling will raise an exception in the communicator, so we connect to the failure handler
        comm.signals.error.connect(self._handle_cancel_config_upload_failed)
        threadpool.start(comm)

    def _handle_cancel_config_upload_failed(self, exception: Exception):
        """Handle failure to cancel the config upload."""
        self._set_busy_wrapper(enabled=False)

        validation_results = self.device_table_view.get_validation_results()
        devices_to_update = []
        for config, config_status, connection_status in validation_results.values():
            devices_to_update.append(
                (config, config_status, ConnectionStatus.UNKNOWN.value, "Upload Cancelled")
            )
        # Rerun validation of all devices after cancellation
        self.device_table_view.update_multiple_device_validations(devices_to_update)
        self.ophyd_test_view.change_device_configs(
            [cfg for cfg, _, _, _ in devices_to_update], added=True, skip_validation=False
        )
        # Config is in sync with BEC, so we update the state
        self.device_table_view.device_config_in_sync_with_redis.emit(False)

    def _handle_push_complete_to_communicator(self):
        """Handle completion of the config push to Redis."""
        self._set_busy_wrapper(enabled=False)

    def _handle_exception_from_communicator(self, exception: Exception):
        """Handle exceptions from the config communicator."""
        QMessageBox.critical(
            self,
            "Error Uploading Config",
            f"An error occurred while uploading the configuration to BEC Server:\n{str(exception)}",
        )
        self._set_busy_wrapper(enabled=False)

    @SafeSlot()
    def _save_to_disk_action(self):
        """Action for the 'save_to_disk' action to save the current config to disk."""
        # Check if plugin repo is installed...
        try:
            config_path = self._get_recovery_config_path()
        except ValueError:
            # Get the recovery config path as fallback
            config_path = os.path.abspath(os.path.expanduser("~"))
            logger.warning(f"Failed to find recovery config path, fallback to: {config_path}")

        # Implement the file loading logic here
        file_path = self._get_file_path(config_path, "save_file")
        if file_path:
            config = {cfg.pop("name"): cfg for cfg in self.device_table_view.get_device_config()}
            if os.path.exists(file_path):
                reply = _yes_no_question(
                    self,
                    "Overwrite File",
                    f"The file '{file_path}' already exists. Do you want to overwrite it?",
                )
                if reply != QMessageBox.StandardButton.Yes:
                    return
            with open(file_path, "w") as file:
                file.write(yaml.dump(config))

    # Table actions
    @SafeSlot()
    def _reset_composed_view(self):
        """Action for the 'reset_composed_view' action to reset the composed view."""
        reply = _yes_no_question(
            self,
            "Clear View",
            "You are about to clear the current composed config view, please confirm...",
        )
        if reply == QMessageBox.StandardButton.Yes:
            self.device_table_view.clear_device_configs()

    @SafeSlot(dict)
    def _edit_device_action(self, device_config: dict):
        """Action to edit a selected device configuration."""
        dialog = DeviceFormDialog(parent=self, add_btn_text="Apply Changes")
        dialog.accepted_data.connect(self._update_device_to_table_from_dialog)
        dialog.set_device_config(device_config)
        dialog.open()

    @SafeSlot()
    def _add_device_action(self):
        """Action for the 'add_device' action to add a new device."""
        dialog = DeviceFormDialog(parent=self, add_btn_text="Add Device")
        dialog.accepted_data.connect(self._add_to_table_from_dialog)
        dialog.open()

    @SafeSlot(dict, int, int, str, str)
    def _update_device_to_table_from_dialog(
        self,
        data: dict,
        config_status: int,
        connection_status: int,
        msg: str,
        old_device_name: str = "",
    ):
        if old_device_name and old_device_name != data.get("name", ""):
            self.device_table_view.remove_device(old_device_name)
        self._add_to_table_from_dialog(data, config_status, connection_status, msg, old_device_name)

    @SafeSlot(dict, int, int, str, str)
    def _add_to_table_from_dialog(
        self,
        data: dict,
        config_status: int,
        connection_status: int,
        msg: str,
        old_device_name: str = "",
    ):
        if connection_status == ConnectionStatus.UNKNOWN.value:
            self.device_table_view.update_device_configs([data], skip_validation=False)
        else:  # Connection status was tested in dialog
            # If device is connected, we remove it from the ophyd validation view
            self.device_table_view.update_device_configs([data], skip_validation=True)
            # Update validation status in device table view and ophyd validation view
            self.ophyd_test_view._on_device_test_completed(
                data, config_status, connection_status, msg
            )

    @SafeSlot()
    def _remove_device_action(self):
        """Action for the 'remove_device' action to remove a device."""
        configs = self.device_table_view.get_selected_device_configs()
        if not configs:
            QMessageBox.warning(
                self, "No devices selected", "Please select devices from the table to remove."
            )
            return
        if self.device_table_view._remove_configs_dialog([cfg["name"] for cfg in configs]):
            self.device_table_view.remove_device_configs(configs)

    @SafeSlot(dict, int, int, str, str)
    def _ophyd_test_item_clicked_cb(
        self, device_config: dict, config_status: int, connection_status: int, msg: str, md_msg: str
    ) -> None:
        self.validation_results.setMarkdown(md_msg)

    def _get_recovery_config_path(self) -> str:
        """Get the recovery config path from the log_writer config."""
        # pylint: disable=protected-access
        log_writer_config = self.client._service_config.config.get("log_writer", {})
        writer = DeviceConfigWriter(service_config=log_writer_config)
        return os.path.abspath(os.path.expanduser(writer.get_recovery_directory()))


if __name__ == "__main__":  # pragma: no cover
    import sys

    from qtpy.QtWidgets import QApplication

    from bec_widgets.widgets.utility.visual.dark_mode_button.dark_mode_button import DarkModeButton

    app = QApplication(sys.argv)
    w = QWidget()
    l = QVBoxLayout()
    w.setLayout(l)
    apply_theme("dark")
    button = DarkModeButton()
    l.addWidget(button)
    device_manager_view = DeviceManagerDisplayWidget()
    l.addWidget(device_manager_view)
    w.show()
    w.setWindowTitle("Device Manager View")
    screen = app.primaryScreen()
    screen_geometry = screen.availableGeometry()
    screen_width = screen_geometry.width()
    screen_height = screen_geometry.height()
    # 70% of screen height, keep 16:9 ratio
    height = int(screen_height * 0.9)
    width = int(height * (16 / 9))

    # If width exceeds screen width, scale down
    if width > screen_width * 0.9:
        width = int(screen_width * 0.9)
        height = int(width / (16 / 9))

    w.resize(width, height)
    sys.exit(app.exec_())



"""Mat-card like widget to display experiment details. Optionally, a button on the bottom which the user can click to trigger the selection of the experiment."""

from bec_lib.messages import ExperimentInfoMessage
from qtpy.QtCore import Qt, Signal
from qtpy.QtWidgets import (
    QFrame,
    QGraphicsDropShadowEffect,
    QGridLayout,
    QGroupBox,
    QHBoxLayout,
    QLabel,
    QPushButton,
    QSizePolicy,
    QVBoxLayout,
    QWidget,
)

from bec_widgets.utils.bec_widget import BECWidget
from bec_widgets.utils.colors import get_theme_palette
from bec_widgets.utils.round_frame import RoundedFrame
from bec_widgets.widgets.services.bec_atlas_admin_view.experiment_selection.utils import (
    format_name,
    format_schedule,
)
from bec_widgets.widgets.utility.visual.dark_mode_button.dark_mode_button import DarkModeButton


class BorderLessLabel(QLabel):
    """A QLabel that does not show any border, even when stylesheets try to apply one."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setStyleSheet("border: none;")
        self.setSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Preferred)


class ExperimentMatCard(BECWidget, QWidget):

    RPC = False

    experiment_selected = Signal(dict)

    def __init__(
        self,
        parent=None,
        show_activate_button: bool = True,
        button_text: str = "Activate",
        title: str = "Next Experiment",
        **kwargs,
    ):
        super().__init__(parent=parent, theme_update=True, **kwargs)

        layout = QVBoxLayout(self)
        layout.setContentsMargins(12, 8, 12, 8)
        self.experiment_info = {}
        self._abstract_text = ""

        # Add card frame with shadow and custom styling
        self._card_frame = QFrame(parent=self)
        layout = QVBoxLayout(self._card_frame)
        self._card_frame.setSizePolicy(QSizePolicy.Preferred, QSizePolicy.Preferred)
        palette = get_theme_palette()
        self._card_frame.setStyleSheet(f"""
                QFrame {{
                    border: 1px solid {palette.mid().color().name()};
                    background: {palette.base().color().name()};
                }}
            """)
        shadow = QGraphicsDropShadowEffect(self._card_frame)
        shadow.setBlurRadius(18)
        shadow.setOffset(0, 4)
        shadow.setColor(palette.shadow().color())
        self._card_frame.setGraphicsEffect(shadow)

        self._group_box = QGroupBox(self._card_frame)
        self._group_box.setStyleSheet(
            "QGroupBox { border: none; }; QLabel { border: none; padding: 0px; }"
        )
        self._fill_group_box(
            title=title, show_activate_button=show_activate_button, button_text=button_text
        )
        self.apply_theme("light")

    def apply_theme(self, theme: str):
        palette = get_theme_palette()
        self._card_frame.setStyleSheet(f"""
                QFrame {{
                    border: 1px solid {palette.mid().color().name()};
                    background: {palette.base().color().name()};
                }}
            """)
        shadow = self._card_frame.graphicsEffect()
        if isinstance(shadow, QGraphicsDropShadowEffect):
            shadow.setColor(palette.shadow().color())

    def _fill_group_box(
        self, title: str, show_activate_button: bool, button_text: str = "Activate"
    ):
        group_layout = QVBoxLayout(self._group_box)
        group_layout.setContentsMargins(16, 16, 16, 16)
        group_layout.setSpacing(12)

        title_row = QHBoxLayout()
        self._card_title = BorderLessLabel(title, self._group_box)
        self._card_title.setStyleSheet("""
            border: none;
            font-size: 14px;
            font-weight: 600;
            """)

        # Add title row and info button to QH layout, then add it to QV layout
        title_row.addWidget(self._card_title)
        title_row.addStretch(1)
        group_layout.addLayout(title_row)

        self._card_grid = QGridLayout()
        self._card_grid.setHorizontalSpacing(12)
        self._card_grid.setVerticalSpacing(8)
        self._card_grid.setColumnStretch(1, 1)

        self._card_pgroup = BorderLessLabel("-", self._group_box)
        self._card_title_value = BorderLessLabel("-", self._group_box)
        self._card_title_value.setWordWrap(True)
        self._card_title_value.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)
        self._card_name = BorderLessLabel("-", self._group_box)
        self._card_start = BorderLessLabel("-", self._group_box)
        self._card_end = BorderLessLabel("-", self._group_box)

        self._card_row_labels = []

        def _row_label(text):
            label = BorderLessLabel(text, self._group_box)
            self._card_row_labels.append(label)
            return label

        self._card_grid.addWidget(_row_label("Name"), 0, 0)
        self._card_grid.addWidget(self._card_name, 0, 1)
        self._card_grid.addWidget(_row_label("Title"), 1, 0)
        self._card_grid.addWidget(self._card_title_value, 1, 1)
        self._card_grid.addWidget(_row_label("P-group"), 2, 0)
        self._card_grid.addWidget(self._card_pgroup, 2, 1)
        self._card_grid.addWidget(_row_label("Schedule (start)"), 3, 0)
        self._card_grid.addWidget(self._card_start, 3, 1)
        self._card_grid.addWidget(_row_label("Schedule (end)"), 4, 0)
        self._card_grid.addWidget(self._card_end, 4, 1)

        # Add to groupbox
        group_layout.addLayout(self._card_grid)

        # Add abstract field at the bottom of the card.
        self._abstract_label = BorderLessLabel("", self._group_box)
        self._abstract_label.setWordWrap(True)
        self._abstract_label.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)
        group_layout.addWidget(self._abstract_label)

        # Add activate button at the bottom
        self._activate_button = QPushButton(button_text, self._group_box)
        self._activate_button.clicked.connect(self._emit_next_experiment)
        self._activate_button.setSizePolicy(
            QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Preferred
        )
        group_layout.addWidget(self._activate_button, alignment=Qt.AlignmentFlag.AlignHCenter)
        self._activate_button.setVisible(show_activate_button)
        self._activate_button.setEnabled(False)

        self._card_frame.layout().setContentsMargins(12, 12, 12, 12)
        self._card_frame.layout().addWidget(self._group_box)

        card_row = QHBoxLayout()
        card_row.addStretch(0)
        card_row.addWidget(self._card_frame)
        card_row.addStretch(0)

        layout = self.layout()
        layout.addStretch(0)
        layout.addLayout(card_row)
        layout.addStretch(0)

    def _emit_next_experiment(self):
        self.experiment_selected.emit(self.experiment_info)

    def clear_experiment_info(self):
        """
        Clear the experiment information displayed on the card and disable the activate button.
        """
        self._card_pgroup.setText("-")
        self._card_title_value.setText("-")
        self._card_name.setText("-")
        self._card_start.setText("-")
        self._card_end.setText("-")
        self._abstract_text = ""
        self._abstract_label.setText("")
        self.experiment_info = {}
        self._activate_button.setEnabled(False)

    def set_experiment_info(self, info: ExperimentInfoMessage | dict):
        """
        Set the experiment information to display on the card.

        Args:
            info (ExperimentInfoMessage | dict): The experiment information to display. Can be either a
                dictionary or an ExperimentInfoMessage instance.
        """
        if isinstance(info, dict):
            info = ExperimentInfoMessage(**info)

        start, end = format_schedule(info.schedule)
        self._card_pgroup.setText(info.pgroup or "-")
        self._card_title_value.setText(info.title or "-")
        self._card_name.setText(format_name(info))
        self._card_start.setText(start or "-")
        self._card_end.setText(end or "-")
        self._abstract_text = (info.abstract or "").strip()
        self._abstract_label.setText(self._abstract_text if self._abstract_text else "")
        self.experiment_info = info.model_dump()
        self._activate_button.setEnabled(True)

    def set_title(self, title: str):
        """
        Set the title displayed at the top of the card.

        Args:
            title (str): The title text to display.
        """
        self._card_title.setText(title)


if __name__ == "__main__":  # pragma: no cover
    import sys

    from bec_qthemes import apply_theme
    from qtpy.QtWidgets import QApplication

    exp_info = {
        "_id": "p22622",
        "owner_groups": ["admin"],
        "access_groups": ["unx-sls_xda_bs", "p22622"],
        "realm_id": "TestBeamline",
        "proposal": "12345967",
        "title": "Test Experiment for Mat Card Widget",
        "firstname": "John",
        "lastname": "Doe",
        "email": "john.doe@psi.ch",
        "account": "doe_j",
        "pi_firstname": "Jane",
        "pi_lastname": "Smith",
        "pi_email": "jane.smith@psi.ch",
        "pi_account": "smith_j",
        "eaccount": "e22622",
        "pgroup": "p22622",
        "abstract": "This is a test abstract for the experiment mat card widget. It should be long enough to test text wrapping and display in the card. The abstract provides a brief overview of the experiment, its goals, and its significance. This text is meant to simulate a real abstract that might be associated with an experiment in the BEC Atlas system. The card should be able to handle abstracts of varying lengths without any issues, ensuring that the user can read the full abstract even if it is quite long.",
        "schedule": [{"start": "01/01/2025 08:00:00", "end": "03/01/2025 18:00:00"}],
        "proposal_submitted": "15/12/2024",
        "proposal_expire": "31/12/2025",
        "proposal_status": "Scheduled",
        "delta_last_schedule": 30,
        "mainproposal": "",
    }

    app = QApplication(sys.argv)

    apply_theme("dark")
    w = QWidget()
    l = QVBoxLayout(w)
    button = DarkModeButton()
    widget = ExperimentMatCard()
    widget.set_experiment_info(exp_info)
    widget.set_title("Scheduled Experiment")
    l.addWidget(button)
    l.addWidget(widget)
    w.resize(w.sizeHint())
    w.show()
    sys.exit(app.exec())



from bec_lib.callback_handler import EventType
from bec_lib.device import ReadoutPriority
from qtpy.QtCore import QSize, Signal, Slot
from qtpy.QtWidgets import QComboBox, QSizePolicy

from bec_widgets.utils.colors import get_accent_colors
from bec_widgets.utils.error_popups import SafeProperty
from bec_widgets.widgets.control.device_input.base_classes.device_input_base import (
    BECDeviceFilter,
    DeviceInputBase,
    DeviceInputConfig,
)


class DeviceComboBox(DeviceInputBase, QComboBox):
    """
    Combobox widget for device input with autocomplete for device names.

    Args:
        parent: Parent widget.
        client: BEC client object.
        config: Device input configuration.
        gui_id: GUI ID.
        device_filter: Device filter, name of the device class from BECDeviceFilter and BECReadoutPriority. Check DeviceInputBase for more details.
        readout_priority_filter: Readout priority filter, name of the readout priority class from BECDeviceFilter and BECReadoutPriority. Check DeviceInputBase for more details.
        available_devices: List of available devices, if passed, it sets apply filters to false and device/readout priority filters will not be applied.
        default: Default device name.
        arg_name: Argument name, can be used for the other widgets which has to call some other function in bec using correct argument names.
        signal_class_filter: List of signal classes to filter the devices by. Only devices with signals of these classes will be shown.
    """

    ICON_NAME = "list_alt"
    PLUGIN = True
    RPC = False

    device_selected = Signal(str)
    device_reset = Signal()
    device_config_update = Signal()

    def __init__(
        self,
        parent=None,
        client=None,
        config: DeviceInputConfig = None,
        gui_id: str | None = None,
        device_filter: BECDeviceFilter | list[BECDeviceFilter] | None = None,
        readout_priority_filter: (
            str | ReadoutPriority | list[str] | list[ReadoutPriority] | None
        ) = None,
        available_devices: list[str] | None = None,
        default: str | None = None,
        arg_name: str | None = None,
        signal_class_filter: list[str] | None = None,
        **kwargs,
    ):
        super().__init__(parent=parent, client=client, gui_id=gui_id, config=config, **kwargs)
        if arg_name is not None:
            self.config.arg_name = arg_name
            self.arg_name = arg_name
        self.setSizePolicy(QSizePolicy.Preferred, QSizePolicy.Fixed)
        self.setMinimumSize(QSize(100, 0))
        self._callback_id = None
        self._is_valid_input = False
        self._accent_colors = get_accent_colors()
        self._set_first_element_as_empty = False

        # We do not consider the config that is passed here, this produced problems
        # with QtDesigner, since config and input arguments may differ and resolve properly
        # Implementing this logic and config recoverage is postponed.
        # Set available devices if passed
        if available_devices is not None:
            self.set_available_devices(available_devices)
        # Set readout priority filter default is all
        if readout_priority_filter is not None:
            self.set_readout_priority_filter(readout_priority_filter)
        else:
            self.set_readout_priority_filter(
                [
                    ReadoutPriority.MONITORED,
                    ReadoutPriority.BASELINE,
                    ReadoutPriority.ASYNC,
                    ReadoutPriority.CONTINUOUS,
                    ReadoutPriority.ON_REQUEST,
                ]
            )
        # Device filter default is None
        if device_filter is not None:
            self.set_device_filter(device_filter)

        if signal_class_filter is not None:
            self.signal_class_filter = signal_class_filter

        # Set default device if passed
        if default is not None:
            self.set_device(default)
        self._callback_id = self.bec_dispatcher.client.callbacks.register(
            EventType.DEVICE_UPDATE, self.on_device_update
        )
        self.device_config_update.connect(self.update_devices_from_filters)
        self.currentTextChanged.connect(self.check_validity)
        self.check_validity(self.currentText())

    @SafeProperty(bool)
    def set_first_element_as_empty(self) -> bool:
        """
        Whether the first element in the combobox should be empty.
        This is useful to allow the user to select a device from the list.
        """
        return self._set_first_element_as_empty

    @set_first_element_as_empty.setter
    def set_first_element_as_empty(self, value: bool) -> None:
        """
        Set whether the first element in the combobox should be empty.
        This is useful to allow the user to select a device from the list.

        Args:
            value (bool): True if the first element should be empty, False otherwise.
        """
        self._set_first_element_as_empty = value
        if self._set_first_element_as_empty:
            self.insertItem(0, "")
            self.setCurrentIndex(0)
        else:
            if self.count() > 0 and self.itemText(0) == "":
                self.removeItem(0)

    def on_device_update(self, action: str, content: dict) -> None:
        """
        Callback for device update events. Triggers the device_update signal.

        Args:
            action (str): The action that triggered the event.
            content (dict): The content of the config update.
        """
        if action in ["add", "remove", "reload"]:
            self.device_config_update.emit()

    def cleanup(self):
        """Cleanup the widget."""
        if self._callback_id is not None:
            self.bec_dispatcher.client.callbacks.remove(self._callback_id)
        super().cleanup()

    def get_current_device(self) -> object:
        """
        Get the current device object based on the current value.

        Returns:
            object: Device object, can be device of type Device, Positioner, Signal or ComputedSignal.
        """
        dev_name = self.currentText()
        return self.get_device_object(dev_name)

    @Slot(str)
    def check_validity(self, input_text: str) -> None:
        """
        Check if the current value is a valid device name.
        """
        if self.validate_device(input_text) is True:
            self._is_valid_input = True
            self.device_selected.emit(input_text)
            self.setStyleSheet("border: 1px solid transparent;")
        else:
            self._is_valid_input = False
            self.device_reset.emit()
            if self.isEnabled():
                self.setStyleSheet("border: 1px solid red;")

    def validate_device(self, device: str) -> bool:  # type: ignore[override]
        """
        Extend validation so that preview‑signal pseudo‑devices (labels like
        ``"eiger_preview"``) are accepted as valid choices.

        The validation run only on device not on the preview‑signal.

        Args:
            device: The text currently entered/selected.

        Returns:
            True if the device is a genuine BEC device *or* one of the
            whitelisted preview‑signal entries.
        """
        idx = self.findText(device)
        if idx >= 0 and isinstance(self.itemData(idx), tuple):
            device = self.itemData(idx)[0]  # type: ignore[assignment]
        return super().validate_device(device)

    @property
    def is_valid_input(self) -> bool:
        """Whether the current text represents a valid device selection."""
        return self._is_valid_input


if __name__ == "__main__":  # pragma: no cover
    # pylint: disable=import-outside-toplevel
    from qtpy.QtWidgets import (
        QApplication,
        QCheckBox,
        QHBoxLayout,
        QLabel,
        QLineEdit,
        QVBoxLayout,
        QWidget,
    )

    from bec_widgets.utils.colors import apply_theme

    app = QApplication([])
    apply_theme("dark")
    widget = QWidget()
    widget.setWindowTitle("DeviceComboBox demo")
    layout = QVBoxLayout(widget)

    layout.addWidget(QLabel("Device filter controls"))
    controls = QHBoxLayout()
    layout.addLayout(controls)

    class_input = QLineEdit()
    class_input.setPlaceholderText("signal_class_filter (comma-separated), e.g. AsyncSignal")
    controls.addWidget(class_input)

    filter_device = QCheckBox("Device")
    filter_positioner = QCheckBox("Positioner")
    filter_signal = QCheckBox("Signal")
    filter_computed = QCheckBox("ComputedSignal")
    controls.addWidget(filter_device)
    controls.addWidget(filter_positioner)
    controls.addWidget(filter_signal)
    controls.addWidget(filter_computed)

    combo = DeviceComboBox()
    combo.set_first_element_as_empty = True
    layout.addWidget(combo)

    def _apply_filters():
        raw = class_input.text().strip()
        if raw:
            combo.signal_class_filter = [entry.strip() for entry in raw.split(",") if entry.strip()]
        else:
            combo.signal_class_filter = []
        combo.filter_to_device = filter_device.isChecked()
        combo.filter_to_positioner = filter_positioner.isChecked()
        combo.filter_to_signal = filter_signal.isChecked()
        combo.filter_to_computed_signal = filter_computed.isChecked()

    class_input.textChanged.connect(_apply_filters)
    filter_device.toggled.connect(_apply_filters)
    filter_positioner.toggled.connect(_apply_filters)
    filter_signal.toggled.connect(_apply_filters)
    filter_computed.toggled.connect(_apply_filters)
    _apply_filters()

    widget.show()
    app.exec_()



from bec_qthemes import material_icon
from qtpy.QtCore import Qt
from qtpy.QtWidgets import QHBoxLayout, QMessageBox, QPushButton, QToolButton, QWidget

from bec_widgets.utils.bec_widget import BECWidget
from bec_widgets.utils.error_popups import SafeSlot


class ResetButton(BECWidget, QWidget):
    """A button that resets the scan queue."""

    PLUGIN = True
    ICON_NAME = "restart_alt"
    RPC = False

    def __init__(self, parent=None, client=None, config=None, gui_id=None, toolbar=False, **kwargs):
        super().__init__(parent=parent, client=client, gui_id=gui_id, config=config, **kwargs)
        self.get_bec_shortcuts()

        self.layout = QHBoxLayout(self)
        self.layout.setSpacing(0)
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.layout.setAlignment(Qt.AlignmentFlag.AlignVCenter)

        if toolbar:
            icon = material_icon(
                "restart_alt", color="#F19E39", filled=True, convert_to_pixmap=False
            )
            self.button = QToolButton(icon=icon)
            self.button.setToolTip("Reset the scan queue")
        else:
            self.button = QPushButton()
            self.button.setText("Reset Queue")
            self.button.setStyleSheet(
                "background-color:  #F19E39; color: white; font-weight: bold; font-size: 12px;"
            )
        self.button.clicked.connect(self.confirm_reset_queue)

        self.layout.addWidget(self.button)

    @SafeSlot()
    def confirm_reset_queue(self):
        """Prompt the user to confirm the queue reset."""
        msg_box = QMessageBox()
        msg_box.setIcon(QMessageBox.Warning)
        msg_box.setWindowTitle("Confirm Reset")
        msg_box.setText(
            "Are you sure you want to reset the scan queue? This action cannot be undone."
        )
        msg_box.setStandardButtons(QMessageBox.Yes | QMessageBox.No)
        msg_box.setDefaultButton(QMessageBox.No)

        if msg_box.exec_() == QMessageBox.Yes:
            self.reset_queue()

    @SafeSlot()
    def reset_queue(self):
        """Reset the scan queue."""
        self.queue.request_queue_reset()



from __future__ import annotations

import enum
import os
import time
from typing import Literal

import numpy as np
from bec_lib import messages
from bec_lib.endpoints import MessageEndpoints
from bec_lib.logger import bec_logger
from qtpy.QtCore import QObject, QTimer, Signal
from qtpy.QtWidgets import QVBoxLayout, QWidget

from bec_widgets.utils.bec_widget import BECWidget
from bec_widgets.utils.error_popups import SafeProperty, SafeSlot
from bec_widgets.utils.ui_loader import UILoader
from bec_widgets.widgets.progress.bec_progressbar.bec_progressbar import ProgressState

logger = bec_logger.logger


class ProgressSource(enum.Enum):
    """
    Enum to define the source of the progress.
    """

    SCAN_PROGRESS = "scan_progress"
    DEVICE_PROGRESS = "device_progress"


class ProgressTask(QObject):
    """
    Class to store progress information.
    Inspired by https://github.com/Textualize/rich/blob/master/rich/progress.py
    """

    def __init__(self, parent: QWidget, value: float = 0, max_value: float = 0, done: bool = False):
        super().__init__(parent=parent)
        self.start_time = time.time()
        self.done = done
        self.value = value
        self.max_value = max_value
        self._elapsed_time = 0

        self.timer = QTimer(self)
        self.timer.timeout.connect(self.update_elapsed_time)
        self.timer.start(100)  # update the elapsed time every 100 ms

    def update(self, value: float, max_value: float, done: bool = False):
        """
        Update the progress.
        """
        self.max_value = max_value
        self.done = done
        self.value = value
        if done:
            self.timer.stop()

    def update_elapsed_time(self):
        """
        Update the time estimates. This is called every 100 ms by a QTimer.
        """
        self._elapsed_time += 0.1

    @property
    def percentage(self) -> float:
        """float: Get progress of task as a percentage. If a None total was set, returns 0"""
        if not self.max_value:
            return 0.0
        completed = (self.value / self.max_value) * 100.0
        completed = min(100.0, max(0.0, completed))
        return completed

    @property
    def speed(self) -> float:
        """Get the estimated speed in steps per second."""
        if self._elapsed_time == 0:
            return 0.0

        return self.value / self._elapsed_time

    @property
    def frequency(self) -> float:
        """Get the estimated frequency in steps per second."""
        if self.speed == 0:
            return 0.0
        return 1 / self.speed

    @property
    def time_elapsed(self) -> str:
        # format the elapsed time to a string in the format HH:MM:SS
        return self._format_time(int(self._elapsed_time))

    @property
    def remaining(self) -> float:
        """Get the estimated remaining steps."""
        if self.done:
            return 0.0
        remaining = self.max_value - self.value
        return remaining

    @property
    def time_remaining(self) -> str:
        """
        Get the estimated remaining time in the format HH:MM:SS.
        """
        if self.done or not self.speed or not self.remaining:
            return self._format_time(0)
        estimate = int(np.round(self.remaining / self.speed))

        return self._format_time(estimate)

    def _format_time(self, seconds: float) -> str:
        """
        Format the time in seconds to a string in the format HH:MM:SS.
        """
        return f"{seconds // 3600:02}:{(seconds // 60) % 60:02}:{seconds % 60:02}"


class ScanProgressBar(BECWidget, QWidget):
    """
    Widget to display a progress bar that is hooked up to the scan progress of a scan.
    If you want to manually set the progress, it is recommended to use the BECProgressbar or QProgressbar directly.
    """

    ICON_NAME = "timelapse"
    PLUGIN = True
    progress_started = Signal()
    progress_finished = Signal()

    def __init__(
        self, parent=None, client=None, config=None, gui_id=None, one_line_design=False, **kwargs
    ):
        super().__init__(parent=parent, client=client, config=config, gui_id=gui_id, **kwargs)

        self.get_bec_shortcuts()
        ui_file = os.path.join(
            os.path.dirname(__file__),
            "scan_progressbar_one_line.ui" if one_line_design else "scan_progressbar.ui",
        )
        self.ui = UILoader(self).loader(ui_file)
        self.layout = QVBoxLayout(self)
        self.layout.setSpacing(0)
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.layout.addWidget(self.ui)
        self.setLayout(self.layout)
        self.progressbar = self.ui.progressbar
        self._show_elapsed_time = self.ui.elapsed_time_label.isVisible()
        self._show_remaining_time = self.ui.remaining_time_label.isVisible()
        self._show_source_label = self.ui.source_label.isVisible()

        self.connect_to_queue()
        self._progress_source = None
        self.task = None
        self.scan_number = None
        self.progress_started.connect(lambda: print("Scan progress started"))

    def connect_to_queue(self):
        """
        Connect to the queue status signal.
        """
        self.bec_dispatcher.connect_slot(self.on_queue_update, MessageEndpoints.scan_queue_status())

    def set_progress_source(self, source: ProgressSource, device=None):
        """
        Set the source of the progress.
        """
        if self._progress_source == source:
            self.update_source_label(source, device=device)
            return
        if self._progress_source is not None:
            self.bec_dispatcher.disconnect_slot(
                self.on_progress_update,
                (
                    MessageEndpoints.scan_progress()
                    if self._progress_source == ProgressSource.SCAN_PROGRESS
                    else MessageEndpoints.device_progress(device=device)
                ),
            )
        self._progress_source = source
        self.bec_dispatcher.connect_slot(
            self.on_progress_update,
            (
                MessageEndpoints.scan_progress()
                if source == ProgressSource.SCAN_PROGRESS
                else MessageEndpoints.device_progress(device=device)
            ),
        )
        self.update_source_label(source, device=device)
        # self.progress_started.emit()

    def update_source_label(self, source: ProgressSource, device=None):
        scan_text = f"Scan {self.scan_number}" if self.scan_number is not None else "Scan"
        text = scan_text if source == ProgressSource.SCAN_PROGRESS else f"Device {device}"
        logger.info(f"Set progress source to {text}")
        self.ui.source_label.setText(text)

    @SafeSlot(dict, dict)
    def on_progress_update(self, msg_content: dict, metadata: dict):
        """
        Update the progress bar based on the progress message.
        """
        value = msg_content["value"]
        max_value = msg_content.get("max_value", 100)
        done = msg_content.get("done", False)
        status: Literal["open", "paused", "aborted", "halted", "closed"] = metadata.get(
            "status", "open"
        )

        if self.task is None:
            return
        self.task.update(value, max_value, done)

        self.update_labels()

        self.progressbar.set_maximum(self.task.max_value)
        self.progressbar.state = ProgressState.from_bec_status(status)
        self.progressbar.set_value(self.task.value)

        if done:
            self.task = None
            self.progress_finished.emit()
            return

    @SafeProperty(bool)
    def show_elapsed_time(self):
        return self._show_elapsed_time

    @show_elapsed_time.setter
    def show_elapsed_time(self, value):
        self._show_elapsed_time = value
        self.ui.elapsed_time_label.setVisible(value)
        if hasattr(self.ui, "dash"):
            self.ui.dash.setVisible(value)

    @SafeProperty(bool)
    def show_remaining_time(self):
        return self._show_remaining_time

    @show_remaining_time.setter
    def show_remaining_time(self, value):
        self._show_remaining_time = value
        self.ui.remaining_time_label.setVisible(value)
        if hasattr(self.ui, "dash"):
            self.ui.dash.setVisible(value)

    @SafeProperty(bool)
    def show_source_label(self):
        return self._show_source_label

    @show_source_label.setter
    def show_source_label(self, value):
        self._show_source_label = value
        self.ui.source_label.setVisible(value)

    def update_labels(self):
        """
        Update the labels based on the progress task.
        """
        if self.task is None:
            return

        self.ui.elapsed_time_label.setText(self.task.time_elapsed)
        self.ui.remaining_time_label.setText(self.task.time_remaining)

    @SafeSlot(dict, dict, verify_sender=True)
    def on_queue_update(self, msg_content, metadata):
        """
        Update the progress bar based on the queue status.
        """
        if not "queue" in msg_content:
            return
        if "primary" not in msg_content["queue"]:
            return
        if (primary_queue := msg_content.get("queue").get("primary")) is None:
            return
        if not isinstance(primary_queue, messages.ScanQueueStatus):
            return
        primary_queue_info = primary_queue.info
        if len(primary_queue_info) == 0:
            return
        scan_info = primary_queue_info[0]
        if scan_info is None:
            return
        if scan_info.status.lower() == "running" and self.task is None:
            self.task = ProgressTask(parent=self)
            self.progress_started.emit()

        active_request_block = scan_info.active_request_block
        if active_request_block is None:
            return

        self.scan_number = active_request_block.scan_number
        report_instructions = active_request_block.report_instructions
        if not report_instructions:
            return

        # for now, let's just use the first instruction
        instruction = report_instructions[0]

        if "scan_progress" in instruction:
            self.set_progress_source(ProgressSource.SCAN_PROGRESS)
        elif "device_progress" in instruction:
            device = instruction["device_progress"][0]
            self.set_progress_source(ProgressSource.DEVICE_PROGRESS, device=device)

    def cleanup(self):
        if self.task is not None:
            self.task.timer.stop()
            self.close()
            self.deleteLater()
        if self._progress_source is not None:
            self.bec_dispatcher.disconnect_slot(
                self.on_progress_update,
                (
                    MessageEndpoints.scan_progress()
                    if self._progress_source == ProgressSource.SCAN_PROGRESS
                    else MessageEndpoints.device_progress(device=self._progress_source.value)
                ),
            )
        self.progressbar.close()
        self.progressbar.deleteLater()
        super().cleanup()


if __name__ == "__main__":  # pragma: no cover
    from qtpy.QtWidgets import QApplication

    bec_logger.disabled_modules = ["bec_lib"]
    app = QApplication([])

    widget = ScanProgressBar()
    widget.show()

    app.exec_()



from bec_ipython_client.main import BECIPythonClient
from qtconsole.inprocess import QtInProcessKernelManager
from qtconsole.manager import QtKernelManager
from qtconsole.rich_jupyter_widget import RichJupyterWidget
from qtpy.QtCore import Qt
from qtpy.QtWidgets import QApplication, QMainWindow


class BECJupyterConsole(RichJupyterWidget):  # pragma: no cover:
    def __init__(self, inprocess: bool = False):
        super().__init__()

        self.inprocess = inprocess
        self.ipyclient = None

        self.kernel_manager, self.kernel_client = self._init_kernel(inprocess=self.inprocess)
        self.set_default_style("linux")
        self._init_bec()

    def _init_kernel(self, inprocess: bool = False, kernel_name: str = "python3"):
        self.inprocess = inprocess
        if inprocess is True:
            print("starting inprocess kernel")
            kernel_manager = QtInProcessKernelManager()
        else:
            kernel_manager = QtKernelManager(kernel_name=kernel_name)
        kernel_manager.start_kernel()
        kernel_client = kernel_manager.client()
        kernel_client.start_channels()
        return kernel_manager, kernel_client

    def _init_bec(self):
        if self.inprocess is True:
            self._init_bec_inprocess()
        else:
            self._init_bec_kernel()

    def _init_bec_inprocess(self):
        self.ipyclient = BECIPythonClient()
        self.ipyclient.start()
        self.kernel_manager.kernel.shell.push(
            {
                "bec": self.ipyclient,
                "dev": self.ipyclient.device_manager.devices,
                "scans": self.ipyclient.scans,
            }
        )

    def _init_bec_kernel(self):
        self.execute("""
            from bec_ipython_client.main import BECIPythonClient
            bec = BECIPythonClient()
            bec.start()
            dev = bec.device_manager.devices if bec else None
            scans = bec.scans if bec else None
            """)

    def _cleanup_bec(self):
        if getattr(self, "ipyclient", None) is not None and self.inprocess is True:
            self.ipyclient.shutdown()
            self.ipyclient = None

    def shutdown_kernel(self):
        """
        Shutdown the Jupyter kernel and clean up resources.
        """
        self._cleanup_bec()
        self.kernel_client.stop_channels()
        self.kernel_manager.shutdown_kernel()
        self.kernel_client = None
        self.kernel_manager = None

    def closeEvent(self, event):
        self.shutdown_kernel()
        event.accept()
        super().closeEvent(event)


class JupyterConsoleWindow(QMainWindow):  # pragma: no cover:
    def __init__(self, inprocess: bool = True, parent=None):
        super().__init__(parent)
        self.console = BECJupyterConsole(inprocess=inprocess)
        self.setCentralWidget(self.console)
        self.setAttribute(Qt.WA_DeleteOnClose, True)

    def closeEvent(self, event):
        # Explicitly close the console so its own closeEvent runs
        if getattr(self, "console", None) is not None:
            self.console.close()
        event.accept()
        super().closeEvent(event)


if __name__ == "__main__":  # pragma: no cover
    import sys

    app = QApplication(sys.argv)
    win = JupyterConsoleWindow(inprocess=True)
    win.show()

    sys.exit(app.exec_())



import sys

from qtpy import QtGui, QtWidgets
from qtpy.QtCore import QPoint, Qt
from qtpy.QtWidgets import (
    QApplication,
    QFrame,
    QHBoxLayout,
    QLabel,
    QProgressBar,
    QVBoxLayout,
    QWidget,
)


class WidgetTooltip(QWidget):
    """Frameless, always-on-top window that behaves like a tooltip."""

    def __init__(self, content: QWidget) -> None:
        super().__init__(
            None,
            Qt.WindowType.ToolTip
            | Qt.WindowType.FramelessWindowHint
            | Qt.WindowType.WindowStaysOnTopHint,
        )
        self.setAttribute(Qt.WidgetAttribute.WA_ShowWithoutActivating)
        self.setAttribute(Qt.WidgetAttribute.WA_TranslucentBackground)
        self.setMouseTracking(True)
        self.content = content

        layout = QVBoxLayout(self)
        layout.setContentsMargins(14, 14, 14, 14)

        self._card = QFrame(self)
        self._card.setObjectName("WidgetTooltipCard")
        card_layout = QVBoxLayout(self._card)
        card_layout.setContentsMargins(12, 10, 12, 10)
        card_layout.addWidget(self.content)

        shadow = QtWidgets.QGraphicsDropShadowEffect(self._card)
        shadow.setBlurRadius(18)
        shadow.setOffset(0, 2)
        shadow.setColor(QtGui.QColor(0, 0, 0, 140))
        self._card.setGraphicsEffect(shadow)

        layout.addWidget(self._card)
        self.apply_theme()
        self.adjustSize()

    def leaveEvent(self, _event) -> None:
        self.hide()

    def apply_theme(self) -> None:
        palette = QApplication.palette()
        base = palette.color(QtGui.QPalette.ColorRole.Base)
        text = palette.color(QtGui.QPalette.ColorRole.Text)
        border = palette.color(QtGui.QPalette.ColorRole.Mid)
        background = QtGui.QColor(base)
        background.setAlpha(242)
        self._card.setStyleSheet(f"""
            QFrame#WidgetTooltipCard {{
                background: {background.name(QtGui.QColor.NameFormat.HexArgb)};
                border: 1px solid {border.name()};
                border-radius: 12px;
            }}
            QFrame#WidgetTooltipCard QLabel {{
                color: {text.name()};
                background: transparent;
            }}
            """)

    def show_above(self, global_pos: QPoint, offset: int = 8) -> None:
        """
        Show the tooltip above a global position, adjusting to stay within screen bounds.

        Args:
            global_pos(QPoint): The global position to show above.
            offset(int, optional): The vertical offset from the global position. Defaults to 8 pixels.
        """
        self.apply_theme()
        self.adjustSize()
        screen = QApplication.screenAt(global_pos) or QApplication.primaryScreen()
        screen_geo = screen.availableGeometry()
        geom = self.geometry()

        x = global_pos.x() - geom.width() // 2
        y = global_pos.y() - geom.height() - offset

        self._navigate_screen_coordinates(screen_geo, geom, x, y)

    def show_near(self, global_pos: QPoint, offset: QPoint | None = None) -> None:
        """
        Show the tooltip near a global position, adjusting to stay within screen bounds.
        By default, it will try to show below and to the right of the position,
        but if that would cause it to go off-screen, it will flip to the other side.

        Args:
            global_pos(QPoint): The global position to show near.
            offset(QPoint, optional): The offset from the global position. Defaults to QPoint(12, 16).
        """

        self.apply_theme()
        self.adjustSize()
        offset = offset or QPoint(12, 16)
        screen = QApplication.screenAt(global_pos) or QApplication.primaryScreen()
        screen_geo = screen.availableGeometry()
        geom = self.geometry()

        x = global_pos.x() + offset.x()
        y = global_pos.y() + offset.y()

        if x + geom.width() > screen_geo.right():
            x = global_pos.x() - geom.width() - abs(offset.x())
        if y + geom.height() > screen_geo.bottom():
            y = global_pos.y() - geom.height() - abs(offset.y())

        self._navigate_screen_coordinates(screen_geo, geom, x, y)

    def _navigate_screen_coordinates(self, screen_geo, geom, x, y):
        x = max(screen_geo.left(), min(x, screen_geo.right() - geom.width()))
        y = max(screen_geo.top(), min(y, screen_geo.bottom() - geom.height()))

        self.move(x, y)
        self.show()
        self.raise_()


class HoverWidget(QWidget):

    def __init__(self, parent: QWidget | None = None, *, simple: QWidget, full: QWidget):
        super().__init__(parent)
        self._simple = simple
        self._full = full
        self._full.setVisible(False)
        self._tooltip = None

        lay = QVBoxLayout(self)
        lay.setContentsMargins(0, 0, 0, 0)
        lay.addWidget(simple)

    def enterEvent(self, event):
        # suppress empty-label tooltips for labels
        if isinstance(self._full, QLabel) and not self._full.text():
            return

        if self._tooltip is None:  # first time only
            self._tooltip = WidgetTooltip(self._full)
            self._full.setVisible(True)

        centre = self.mapToGlobal(self.rect().center())
        self._tooltip.show_above(centre)
        super().enterEvent(event)

    def leaveEvent(self, event):
        if self._tooltip and self._tooltip.isVisible():
            self._tooltip.hide()
        super().leaveEvent(event)

    def close(self):
        if self._tooltip:
            self._tooltip.close()
            self._tooltip.deleteLater()
            self._tooltip = None
        super().close()


################################################################################
# Demo
# Just a simple example to show how the HoverWidget can be used to display
# a tooltip with a full widget inside (two different widgets are used
# for the simple and full versions).
################################################################################


class DemoSimpleWidget(QLabel):  # pragma: no cover
    """A simple widget to be used as a trigger for the tooltip."""

    def __init__(self) -> None:
        super().__init__()
        self.setText("Hover me for a preview!")


class DemoFullWidget(QProgressBar):  # pragma: no cover
    """A full widget to be shown in the tooltip."""

    def __init__(self) -> None:
        super().__init__()
        self.setRange(0, 100)
        self.setValue(75)
        self.setFixedWidth(320)
        self.setFixedHeight(30)


if __name__ == "__main__":  # pragma: no cover
    app = QApplication(sys.argv)

    window = QWidget()
    window.layout = QHBoxLayout(window)
    hover_widget = HoverWidget(simple=DemoSimpleWidget(), full=DemoFullWidget())
    window.layout.addWidget(hover_widget)
    window.show()

    sys.exit(app.exec_())



from textwrap import dedent
from typing import NamedTuple
from uuid import uuid4

from bec_qthemes import material_icon
from qtpy.QtCore import QItemSelection, QSize, Signal
from qtpy.QtWidgets import QFrame, QHBoxLayout, QLabel, QListWidgetItem, QVBoxLayout, QWidget

from bec_widgets.utils.error_popups import SafeSlot
from bec_widgets.utils.expandable_frame import ExpandableGroupFrame
from bec_widgets.widgets.control.device_manager.components._util import SharedSelectionSignal
from bec_widgets.widgets.control.device_manager.components.available_device_resources.available_device_group_ui import (
    Ui_AvailableDeviceGroup,
)
from bec_widgets.widgets.control.device_manager.components.available_device_resources.device_resource_backend import (
    HashableDevice,
)
from bec_widgets.widgets.control.device_manager.components.constants import CONFIG_DATA_ROLE


def _warning_string(spec: HashableDevice):
    name_warning = (
        "Device defined with multiple names! Please check:\n  " + "\n  ".join(spec.names)
        if len(spec.names) > 1
        else ""
    )
    source_warning = (
        "Device found in multiple source files! Please check:\n  " + "\n  ".join(spec._source_files)
        if len(spec._source_files) > 1
        else ""
    )
    return f"{name_warning}{source_warning}"


class _DeviceEntryWidget(QFrame):

    def __init__(self, device_spec: HashableDevice, parent=None, **kwargs):
        super().__init__(parent, **kwargs)
        self._device_spec = device_spec
        self.included: bool = False

        self.setFrameStyle(0)

        self._layout = QVBoxLayout()
        self._layout.setContentsMargins(2, 2, 2, 2)
        self.setLayout(self._layout)

        self.setup_title_layout(device_spec)
        self.check_and_display_warning()

        self.setToolTip(self._rich_text())

    def _rich_text(self):
        return dedent(f"""
        <b><u><h2> {self._device_spec.name}: </h2></u></b>
        <table>
        <tr><td> description: </td><td><i> {self._device_spec.description}  </i></td></tr>
        <tr><td> config:      </td><td><i> {self._device_spec.deviceConfig} </i></td></tr>
        <tr><td> enabled:     </td><td><i> {self._device_spec.enabled}      </i></td></tr>
        <tr><td> read only:   </td><td><i> {self._device_spec.readOnly}     </i></td></tr>
        </table>
        """)

    def setup_title_layout(self, device_spec: HashableDevice):
        self._title_layout = QHBoxLayout()
        self._title_layout.setContentsMargins(0, 0, 0, 0)
        self._title_container = QWidget(parent=self)
        self._title_container.setLayout(self._title_layout)

        self._warning_label = QLabel()
        self._title_layout.addWidget(self._warning_label)

        self.title = QLabel(device_spec.name)
        self.title.setToolTip(device_spec.name)
        self.title.setStyleSheet(self.title_style("#FF0000"))
        self._title_layout.addWidget(self.title)

        self._title_layout.addStretch(1)
        self._layout.addWidget(self._title_container)

    def check_and_display_warning(self):
        if len(self._device_spec.names) == 1 and len(self._device_spec._source_files) == 1:
            self._warning_label.setText("")
            self._warning_label.setToolTip("")
        else:
            self._warning_label.setPixmap(material_icon("warning", size=(12, 12), color="#FFAA00"))
            self._warning_label.setToolTip(_warning_string(self._device_spec))

    @property
    def device_hash(self):
        return hash(self._device_spec)

    def title_style(self, color: str) -> str:
        return f"QLabel {{ color: {color}; font-weight: bold; font-size: 10pt; }}"

    def setTitle(self, text: str):
        self.title.setText(text)

    def set_included(self, included: bool):
        self.included = included
        self.title.setStyleSheet(self.title_style("#00FF00" if included else "#FF0000"))


class _DeviceEntry(NamedTuple):
    list_item: QListWidgetItem
    widget: _DeviceEntryWidget


class AvailableDeviceGroup(ExpandableGroupFrame, Ui_AvailableDeviceGroup):

    selected_devices = Signal(list)

    def __init__(
        self,
        parent=None,
        name: str = "TagGroupTitle",
        data: set[HashableDevice] = set(),
        shared_selection_signal=SharedSelectionSignal(),
        **kwargs,
    ):
        super().__init__(parent=parent, **kwargs)
        self.setupUi(self)

        self._shared_selection_signal = shared_selection_signal
        self._shared_selection_uuid = str(uuid4())
        self._shared_selection_signal.proc.connect(self._handle_shared_selection_signal)
        self.device_list.selectionModel().selectionChanged.connect(self._on_selection_changed)

        self.title_text = name  # type: ignore
        self._mime_data = []
        self._devices: dict[str, _DeviceEntry] = {}
        for device in data:
            self._add_item(device)
        self.device_list.sortItems()
        self.setMinimumSize(self.device_list.sizeHint())
        self._update_num_included()

    def _add_item(self, device: HashableDevice):
        item = QListWidgetItem(self.device_list)
        device_dump = device.model_dump(exclude_defaults=True)
        item.setData(CONFIG_DATA_ROLE, device_dump)
        self._mime_data.append(device_dump)
        widget = _DeviceEntryWidget(device, self)
        item.setSizeHint(QSize(widget.width(), widget.height()))
        self.device_list.setItemWidget(item, widget)
        self.device_list.addItem(item)
        self._devices[device.name] = _DeviceEntry(item, widget)

    def create_mime_data(self):
        return self._mime_data

    def reset_devices_state(self):
        for dev in self._devices.values():
            dev.widget.set_included(False)
        self._update_num_included()

    def set_item_state(self, /, device_hash: int, included: bool):
        for dev in self._devices.values():
            if dev.widget.device_hash == device_hash:
                dev.widget.set_included(included)
        self._update_num_included()

    def _update_num_included(self):
        n_included = sum(int(dev.widget.included) for dev in self._devices.values())
        if n_included == 0:
            color = "#FF0000"
        elif n_included == len(self._devices):
            color = "#00FF00"
        else:
            color = "#FFAA00"
        self.n_included.setText(f"{n_included} / {len(self._devices)}")
        self.n_included.setStyleSheet(f"QLabel {{ color: {color}; }}")

    def sizeHint(self) -> QSize:
        if not getattr(self, "device_list", None) or not self.expanded:
            return super().sizeHint()
        return QSize(
            max(150, self.device_list.viewport().width()),
            self.device_list.sizeHintForRow(0) * self.device_list.count() + 50,
        )

    @SafeSlot(QItemSelection, QItemSelection)
    def _on_selection_changed(self, selected: QItemSelection, deselected: QItemSelection) -> None:
        self._shared_selection_signal.proc.emit(self._shared_selection_uuid)
        config = [dev.as_normal_device().model_dump() for dev in self.get_selection()]
        self.selected_devices.emit(config)

    @SafeSlot(str)
    def _handle_shared_selection_signal(self, uuid: str):
        if uuid != self._shared_selection_uuid:
            self.device_list.clearSelection()

    def resizeEvent(self, event):
        super().resizeEvent(event)
        self.setMinimumHeight(self.sizeHint().height())
        self.setMaximumHeight(self.sizeHint().height())

    def get_selection(self) -> set[HashableDevice]:
        selection = self.device_list.selectedItems()
        widgets = (w.widget for _, w in self._devices.items() if w.list_item in selection)
        return set(w._device_spec for w in widgets)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}: {self.title_text}"


if __name__ == "__main__":
    import sys

    from qtpy.QtWidgets import QApplication

    app = QApplication(sys.argv)
    widget = AvailableDeviceGroup(name="Tag group 1")
    for item in [
        HashableDevice(
            **{
                "name": f"test_device_{i}",
                "deviceClass": "TestDeviceClass",
                "readoutPriority": "baseline",
                "enabled": True,
            }
        )
        for i in range(5)
    ]:
        widget._add_item(item)
    widget._update_num_included()
    widget.show()
    sys.exit(app.exec())



import re
import subprocess
from pathlib import Path

from bec_lib.logger import bec_logger
from bec_lib.plugin_helper import plugin_package_name, plugin_repo_path
from watchdog.events import (
    DirCreatedEvent,
    DirModifiedEvent,
    DirMovedEvent,
    FileCreatedEvent,
    FileModifiedEvent,
    FileMovedEvent,
    FileSystemEvent,
    FileSystemEventHandler,
)
from watchdog.observers import Observer

from bec_widgets.utils.bec_designer import open_designer
from bec_widgets.utils.bec_plugin_helper import get_all_plugin_widgets
from bec_widgets.utils.plugin_utils import get_custom_classes

logger = bec_logger.logger


class RecompileHandler(FileSystemEventHandler):
    def __init__(self, in_file: Path, out_file: Path) -> None:
        super().__init__()
        self.in_file = str(in_file)
        self.out_file = str(out_file)
        self._pyside_import_re = re.compile(r"from PySide6\.(.*) import ")
        self._widget_import_re = re.compile(
            r"^from ([a-zA-Z_]*) import ([a-zA-Z_]*)$", re.MULTILINE
        )
        self._widget_modules = {
            c.name: c.module for c in (get_custom_classes("bec_widgets") + get_all_plugin_widgets())
        }

    def on_created(self, event: DirCreatedEvent | FileCreatedEvent) -> None:
        self.recompile(event)

    def on_modified(self, event: DirModifiedEvent | FileModifiedEvent) -> None:
        self.recompile(event)

    def on_moved(self, event: DirMovedEvent | FileMovedEvent) -> None:
        self.recompile(event)

    def recompile(self, event: FileSystemEvent) -> None:
        if event.src_path == self.in_file or event.dest_path == self.in_file:
            self._recompile()

    def _recompile(self):
        logger.success(".ui file modified, recompiling...")
        code = subprocess.call(
            ["pyside6-uic", "--absolute-imports", self.in_file, "-o", self.out_file]
        )
        logger.success(f"compilation exited with code {code}")
        if code != 0:
            return
        self._add_comment_to_file()
        logger.success("updating imports...")
        self._update_imports()
        logger.success("formatting...")
        code = subprocess.call(
            ["black", "--line-length=100", "--skip-magic-trailing-comma", self.out_file]
        )
        if code != 0:
            logger.error(f"Error while running black on {self.out_file}, code: {code}")
            return
        code = subprocess.call(
            [
                "isort",
                "--line-length=100",
                "--profile=black",
                "--multi-line=3",
                "--trailing-comma",
                self.out_file,
            ]
        )
        if code != 0:
            logger.error(f"Error while running isort on {self.out_file}, code: {code}")
            return
        logger.success("done!")

    def _add_comment_to_file(self):
        with open(self.out_file, "r+") as f:
            initial = f.read()
            f.seek(0)
            f.write(f"# Generated from {self.in_file} by bec-plugin-manager - do not edit! \n")
            f.write(
                "# Use 'bec-plugin-manager edit-ui [widget_name]' to make changes, and this file will be updated accordingly. \n\n"
            )
            f.write(initial)

    def _update_imports(self):
        with open(self.out_file, "r+") as f:
            initial = f.read()
            f.seek(0)
            qtpy_imports = re.sub(
                self._pyside_import_re, lambda ob: f"from qtpy.{ob.group(1)} import ", initial
            )
            print(self._widget_modules)
            print(re.findall(self._widget_import_re, qtpy_imports))
            widget_imports = re.sub(
                self._widget_import_re,
                lambda ob: (
                    f"from {module} import {ob.group(2)}"
                    if (module := self._widget_modules.get(ob.group(2))) is not None
                    else ob.group(1)
                ),
                qtpy_imports,
            )
            f.write(widget_imports)
            f.truncate()


def open_and_watch_ui_editor(widget_name: str):
    logger.info(f"Opening the editor for {widget_name}, and watching")
    repo = Path(plugin_repo_path())
    widget_dir = repo / plugin_package_name() / "bec_widgets" / "widgets" / widget_name
    ui_file = widget_dir / f"{widget_name}.ui"
    ui_outfile = widget_dir / f"{widget_name}_ui.py"

    logger.info(
        f"Opening the editor for {widget_name}, and watching {ui_file} for changes. Whenever you save the file, it will be recompiled to {ui_outfile}"
    )
    recompile_handler = RecompileHandler(ui_file, ui_outfile)
    observer = Observer()
    observer.schedule(recompile_handler, str(ui_file.parent))
    observer.start()
    try:
        open_designer([str(ui_file)])
    finally:
        observer.stop()
        observer.join()
    logger.info("Editing session ended, exiting...")



from bec_lib.config_helper import ConfigHelper
from bec_lib.logger import bec_logger
from bec_lib.messages import ConfigAction
from qtpy.QtCore import QObject, QRunnable, Signal

from bec_widgets.utils.error_popups import SafeSlot

logger = bec_logger.logger


class _CommSignals(QObject):
    error = Signal(Exception)
    done = Signal()


class CommunicateConfigAction(QRunnable):

    def __init__(
        self,
        config_helper: ConfigHelper,
        device: str | None,
        config: dict | None,
        action: ConfigAction,
    ) -> None:
        super().__init__()
        self.config_helper = config_helper
        if action in ["add", "update"] and config is None:
            raise ValueError("Must supply config to add or update a device")
        self.device = device
        self.config = config or {}
        self.action = action
        self.signals = _CommSignals()

    @SafeSlot()
    def run(self):
        try:
            if self.action == "set":
                self._process(
                    {"action": self.action, "config": self.config, "wait_for_response": False}
                )
            elif self.action == "cancel":
                self._process_cancel()
            elif self.action in ["add", "update", "remove"]:
                if (dev_name := self.device or self.config.get("name")) is None:
                    raise ValueError(
                        "Must be updating a device or be supplied a name for a new device"
                    )
                if "deviceConfig" not in self.config or self.action in ["add", "remove"]:
                    self.process_simple_action(dev_name)
                else:
                    # updating an existing device, but need to recreate it for this change
                    self.process_remove_readd(dev_name)
            else:
                raise ValueError(f"action {self.action} is not supported")
        except Exception as e:
            self.signals.error.emit(e)
        else:
            self.signals.done.emit()

    def process_simple_action(self, dev_name: str, action: ConfigAction | None = None):
        req_args = {
            "action": action or self.action,
            "config": {dev_name: self.config},
            "wait_for_response": False,
        }
        self._process(req_args)

    def _process(self, req_args: dict):
        timeout = (
            self.config_helper.suggested_timeout_s(self.config) if self.config is not None else 20
        )
        RID = self.config_helper.send_config_request(**req_args)
        logger.info("Waiting for config reply")
        reply = self.config_helper.wait_for_config_reply(RID, timeout=timeout)
        self.config_helper.handle_update_reply(reply, RID, timeout)
        logger.info("Done updating config!")

    def _process_cancel(self):
        logger.info("Cancelling ongoing configuration operation")
        self.config_helper.send_config_request(
            action="cancel", config=None, wait_for_response=True, timeout_s=10
        )
        logger.info("Done cancelling configuration operation")

    def process_remove_readd(self, dev_name: str):
        logger.info(f"Removing and readding device: {dev_name}")
        self.process_simple_action(dev_name, "remove")
        self.process_simple_action(dev_name, "add")
        logger.info(f"Reinstated {dev_name} successfully!")



from qtpy.QtWidgets import QWidget

from bec_widgets.applications.views.developer_view.developer_widget import DeveloperWidget
from bec_widgets.applications.views.view import ViewBase, ViewTourSteps


class DeveloperView(ViewBase):
    """
    A view for users to write scripts and macros and execute them within the application.
    """

    def __init__(
        self,
        parent: QWidget | None = None,
        content: QWidget | None = None,
        *,
        view_id: str | None = None,
        title: str | None = None,
        **kwargs,
    ):
        super().__init__(parent=parent, content=content, view_id=view_id, title=title, **kwargs)
        self.developer_widget = DeveloperWidget(parent=self)
        self.set_content(self.developer_widget)

    def register_tour_steps(self, guided_tour, main_app) -> ViewTourSteps | None:
        """Register Developer View components with the guided tour.

        Args:
            guided_tour: The GuidedTour instance to register with.
            main_app: The main application instance (for accessing set_current).

        Returns:
            ViewTourSteps | None: Model containing view title and step IDs.
        """
        step_ids = []
        dev_widget = self.developer_widget

        # IDE Toolbar
        def get_ide_toolbar():
            main_app.set_current("developer_view")
            return (dev_widget.toolbar, None)

        step_id = guided_tour.register_widget(
            widget=get_ide_toolbar,
            title="IDE Toolbar",
            text="Quick access to save files, execute scripts, and configure IDE settings. Use the toolbar to manage your code and execution.",
        )
        step_ids.append(step_id)

        # IDE Explorer
        def get_ide_explorer():
            main_app.set_current("developer_view")
            return (dev_widget.explorer_dock.widget(), None)

        step_id = guided_tour.register_widget(
            widget=get_ide_explorer,
            title="File Explorer",
            text="Browse and manage your macro files. Create new files, open existing ones, and organize your scripts.",
        )
        step_ids.append(step_id)

        # IDE Editor
        def get_ide_editor():
            main_app.set_current("developer_view")
            return (dev_widget.monaco_dock.widget(), None)

        step_id = guided_tour.register_widget(
            widget=get_ide_editor,
            title="Code Editor",
            text="Write and edit Python code with syntax highlighting, auto-completion, and signature help. Monaco editor provides a modern coding experience.",
        )
        step_ids.append(step_id)

        # IDE Console
        def get_ide_console():
            main_app.set_current("developer_view")
            return (dev_widget.console_dock.widget(), None)

        step_id = guided_tour.register_widget(
            widget=get_ide_console,
            title="BEC Shell Console",
            text="Interactive Python console with BEC integration. Execute commands, test code snippets, and interact with the BEC system in real-time.",
        )
        step_ids.append(step_id)

        # IDE Plotting Area
        def get_ide_plotting():
            main_app.set_current("developer_view")
            return (dev_widget.plotting_ads, None)

        step_id = guided_tour.register_widget(
            widget=get_ide_plotting,
            title="Plotting Area",
            text="View plots and visualizations generated by your scripts. Arrange multiple plots in a flexible layout.",
        )
        step_ids.append(step_id)

        return ViewTourSteps(view_title="Developer View", step_ids=step_ids)


if __name__ == "__main__":
    import sys

    from bec_qthemes import apply_theme
    from qtpy.QtWidgets import QApplication

    from bec_widgets.applications.main_app import BECMainApp

    app = QApplication(sys.argv)
    apply_theme("dark")

    _app = BECMainApp()
    screen = app.primaryScreen()
    screen_geometry = screen.availableGeometry()
    screen_width = screen_geometry.width()
    screen_height = screen_geometry.height()
    # 70% of screen height, keep 16:9 ratio
    height = int(screen_height * 0.9)
    width = int(height * (16 / 9))

    # If width exceeds screen width, scale down
    if width > screen_width * 0.9:
        width = int(screen_width * 0.9)
        height = int(width / (16 / 9))

    _app.resize(width, height)
    developer_view = DeveloperView()
    _app.add_view(
        icon="code_blocks",
        title="IDE",
        widget=developer_view,
        view_id="developer_view",
        exclusive=True,
    )
    _app.show()
    # developer_view.show()
    # developer_view.setWindowTitle("Developer View")
    # developer_view.resize(1920, 1080)
    # developer_view.set_stretch(horizontal=[1, 3, 2], vertical=[5, 5]) #can be set during runtime
    sys.exit(app.exec_())



import pyqtgraph as pg
from qtpy.QtCore import Property, Qt
from qtpy.QtWidgets import QApplication, QFrame, QHBoxLayout, QVBoxLayout, QWidget

from bec_widgets.widgets.utility.visual.dark_mode_button.dark_mode_button import DarkModeButton


class RoundedFrame(QFrame):
    # TODO this should be removed completely in favor of QSS styling, no time now
    """
    A custom QFrame with rounded corners and optional theme updates.
    The frame can contain any QWidget, however it is mainly designed to wrap PlotWidgets to provide a consistent look and feel with other BEC Widgets.
    """

    def __init__(
        self,
        parent=None,
        content_widget: QWidget = None,
        background_color: str = None,
        orientation: str = "horizontal",
        radius: int = 10,
    ):
        QFrame.__init__(self, parent)

        self.background_color = background_color
        self._radius = radius

        # Apply rounded frame styling
        self.setProperty("skip_settings", True)
        self.setObjectName("roundedFrame")

        # Ensure QSS can paint background/border on this widget
        self.setAttribute(Qt.WA_StyledBackground, True)

        # Create a layout for the frame
        if orientation == "vertical":
            self.layout = QVBoxLayout(self)
            self.layout.setContentsMargins(5, 5, 5, 5)
        else:
            self.layout = QHBoxLayout(self)
            self.layout.setContentsMargins(5, 5, 5, 5)  # Set 5px margin

        # Add the content widget to the layout
        if content_widget:
            self.layout.addWidget(content_widget)

        # Store reference to the content widget
        self.content_widget = content_widget

        # Automatically apply initial styles to the GraphicalLayoutWidget if applicable
        self.apply_plot_widget_style()
        self.update_style()

    def apply_theme(self, theme: str):
        """Deprecated: RoundedFrame no longer handles theme; styling is QSS-driven."""
        self.update_style()

    @Property(int)
    def radius(self):
        """Radius of the rounded corners."""
        return self._radius

    @radius.setter
    def radius(self, value: int):
        self._radius = value
        self.update_style()

    def update_style(self):
        """
        Update the style of the frame based on the background color.
        """
        self.setStyleSheet(f"""
                QFrame#roundedFrame {{
                    border-radius: {self._radius}px;
                }}
            """)
        self.apply_plot_widget_style()

    def apply_plot_widget_style(self, border: str = "none"):
        """
        Let QSS/pyqtgraph handle plot styling; avoid overriding here.
        """
        if isinstance(self.content_widget, pg.GraphicsLayoutWidget):
            self.content_widget.setStyleSheet("")


class ExampleApp(QWidget):  # pragma: no cover
    def __init__(self):
        super().__init__()
        self.setWindowTitle("Rounded Plots Example")

        # Main layout
        layout = QVBoxLayout(self)

        dark_button = DarkModeButton()

        # Create PlotWidgets
        plot1 = pg.GraphicsLayoutWidget()
        plot_item_1 = pg.PlotItem()
        plot_item_1.plot([1, 3, 2, 4, 6, 5], pen="r")
        plot1.plot_item = plot_item_1

        plot2 = pg.GraphicsLayoutWidget()
        plot_item_2 = pg.PlotItem()
        plot_item_2.plot([1, 2, 4, 8, 16, 32], pen="r")
        plot2.plot_item = plot_item_2

        # Add to layout (no RoundedFrame wrapper; QSS styles plots)
        layout.addWidget(dark_button)
        layout.addWidget(plot1)
        layout.addWidget(plot2)

        self.setLayout(layout)

        # Theme flip demo removed; global theming applies automatically


if __name__ == "__main__":  # pragma: no cover
    app = QApplication([])

    window = ExampleApp()
    window.show()

    app.exec()



# from .device_table_view import DeviceTableView
from .device_table.device_table import DeviceTable
from .dm_config_view import DMConfigView
from .dm_docstring_view import DocstringView, docstring_to_markdown
from .ophyd_validation.ophyd_validation import OphydValidation



from qtpy.QtCore import Qt
from qtpy.QtWidgets import QTableWidget


class BECTable(QTableWidget):
    """Table widget with custom keyPressEvent to delete rows with backspace or delete key"""

    def keyPressEvent(self, event) -> None:
        """
        Delete selected rows with backspace or delete key

        Args:
            event: keyPressEvent
        """
        if event.key() in (Qt.Key_Backspace, Qt.Key_Delete):
            selected_ranges = self.selectedRanges()
            for selected_range in selected_ranges:
                for row in range(selected_range.topRow(), selected_range.bottomRow() + 1):
                    self.removeRow(row)
        else:
            super().keyPressEvent(event)



import re


def pascal_to_snake(name: str) -> str:
    """
    Convert PascalCase to snake_case.

    Args:
        name (str): The name to be converted.

    Returns:
        str: The converted name.
    """
    s1 = re.sub(r"([a-z0-9])([A-Z])", r"\1_\2", name)
    s2 = re.sub(r"([A-Z]+)([A-Z][a-z])", r"\1_\2", s1)
    return s2.lower()


def sanitize_namespace(namespace: str | None) -> str | None:
    """
    Clean user-provided namespace labels for filesystem compatibility.

    Args:
        namespace (str | None): Arbitrary namespace identifier supplied by the caller.

    Returns:
        str | None: Sanitized namespace containing only safe characters, or ``None``
            when the input is empty.
    """
    if not namespace:
        return None
    ns = namespace.strip()
    if not ns:
        return None
    return re.sub(r"[^0-9A-Za-z._-]+", "_", ns)



from __future__ import annotations

from collections import defaultdict
from typing import Literal

import numpy as np
from bec_lib import bec_logger
from bec_lib.endpoints import MessageEndpoints
from pydantic import BaseModel, Field, field_validator
from qtpy.QtCore import QTimer
from qtpy.QtWidgets import QWidget

from bec_widgets.utils import ConnectionConfig
from bec_widgets.utils.colors import Colors, apply_theme
from bec_widgets.utils.error_popups import SafeProperty, SafeSlot
from bec_widgets.widgets.plots.image.image_base import ImageBase
from bec_widgets.widgets.plots.image.image_item import ImageItem
from bec_widgets.widgets.plots.image.toolbar_components.device_selection import (
    DeviceSelection,
    DeviceSelectionConnection,
    device_selection_bundle,
)
from bec_widgets.widgets.plots.plot_base import PlotBase

logger = bec_logger.logger


# noinspection PyDataclass
class ImageConfig(ConnectionConfig):
    color_map: str = Field(
        "plasma", description="The colormap  of the figure widget.", validate_default=True
    )
    color_bar: Literal["full", "simple"] | None = Field(
        None, description="The type of the color bar."
    )
    lock_aspect_ratio: bool = Field(
        False, description="Whether to lock the aspect ratio of the image."
    )

    model_config: dict = {"validate_assignment": True}
    _validate_color_map = field_validator("color_map")(Colors.validate_color_map)


class ImageLayerConfig(BaseModel):
    device: str = Field("", description="The device name to monitor.")
    signal: str = Field("", description="The signal/entry name to monitor on the device.")
    monitor_type: Literal["1d", "2d"] | None = Field(None, description="The type of monitor.")
    source: Literal["device_monitor_1d", "device_monitor_2d"] | None = Field(
        None, description="The source of the image data."
    )
    async_signal_name: str | None = Field(
        None, description="Async signal name (obj_name) used for async endpoints."
    )
    connection_status: Literal["connected", "disconnected", "error"] = Field(
        "disconnected", description="Current connection status."
    )
    connection_error: str | None = Field(None, description="Last connection error, if any.")


class Image(ImageBase):
    """
    Image widget for displaying 2D data.
    """

    PLUGIN = True
    RPC = True
    ICON_NAME = "image"
    USER_ACCESS = [
        *PlotBase.USER_ACCESS,
        # ImageView Specific Settings
        "color_map",
        "color_map.setter",
        "v_range",
        "v_range.setter",
        "v_min",
        "v_min.setter",
        "v_max",
        "v_max.setter",
        "autorange",
        "autorange.setter",
        "autorange_mode",
        "autorange_mode.setter",
        "device",
        "device.setter",
        "signal",
        "signal.setter",
        "enable_colorbar",
        "enable_simple_colorbar",
        "enable_simple_colorbar.setter",
        "enable_full_colorbar",
        "enable_full_colorbar.setter",
        "fft",
        "fft.setter",
        "log",
        "log.setter",
        "num_rotation_90",
        "num_rotation_90.setter",
        "transpose",
        "transpose.setter",
        "image",
        "main_image",
        "add_roi",
        "remove_roi",
        "rois",
    ]

    SUPPORTED_SIGNALS = ["AsyncSignal", "AsyncMultiSignal", "DynamicSignal"]

    def __init__(
        self,
        parent: QWidget | None = None,
        config: ImageConfig | None = None,
        client=None,
        gui_id: str | None = None,
        popups: bool = True,
        **kwargs,
    ):
        if config is None:
            config = ImageConfig(widget_class=self.__class__.__name__)
        self.gui_id = config.gui_id
        self.subscriptions: defaultdict[str, ImageLayerConfig] = defaultdict(ImageLayerConfig)
        # Store signal configs separately (not serialized to QSettings)
        self._signal_configs: dict[str, dict] = {}

        super().__init__(
            parent=parent, config=config, client=client, gui_id=gui_id, popups=popups, **kwargs
        )
        self._device_selection_updating = False
        self._autorange_on_next_update = False
        self._init_toolbar_image()
        self.layer_removed.connect(self._on_layer_removed)
        self.old_scan_id = None
        self.scan_id = None
        self.async_update = False
        self.bec_dispatcher.connect_slot(self.on_scan_status, MessageEndpoints.scan_status())
        self.bec_dispatcher.connect_slot(self.on_scan_progress, MessageEndpoints.scan_progress())

    @property
    def _config(self) -> ImageLayerConfig:
        """Helper property to access the main layer config."""
        return self.subscriptions["main"]

    ##################################
    ### Toolbar Initialization
    ##################################

    def _init_toolbar_image(self):
        """
        Initializes the toolbar for the image widget.
        """
        self.toolbar.add_bundle(
            device_selection_bundle(self.toolbar.components, client=self.client)
        )
        self.toolbar.connect_bundle(
            "device_selection",
            DeviceSelectionConnection(self.toolbar.components, target_widget=self),
        )

        crosshair_bundle = self.toolbar.get_bundle("image_crosshair")
        crosshair_bundle.add_action("image_autorange")
        crosshair_bundle.add_action("image_colorbar_switch")

        self.toolbar.show_bundles(
            [
                "device_selection",
                "plot_export",
                "mouse_interaction",
                "image_crosshair",
                "image_processing",
                "axis_popup",
            ]
        )

        QTimer.singleShot(0, self._adjust_and_connect)

    def _adjust_and_connect(self):
        """
        Sync the device selection toolbar with current properties.
        Has to be done with QTimer.singleShot to ensure the UI is fully initialized, needed for testing.

        Note: DeviceComboBox and SignalComboBox auto-populate themselves, no manual population needed.
        """
        self._sync_device_selection()

    @SafeSlot()
    def on_device_selection_changed(self, _):
        """
        Called when device or signal selection changes in the toolbar.
        This reads from the toolbar and updates the widget properties.
        """
        if self._device_selection_updating:
            return

        self._device_selection_updating = True
        try:
            try:
                action = self.toolbar.components.get_action("device_selection")
            except Exception:
                return

            if action is None:
                return

            device_selection: DeviceSelection = action.widget
            device = device_selection.device_combo_box.currentText()
            signal_text = device_selection.signal_combo_box.currentText()

            if not device:
                self.device = ""
                return
            if not device_selection.device_combo_box.is_valid_input:
                return

            if not device_selection.signal_combo_box.is_valid_input:
                if self._config.signal:
                    self.signal = ""
                if device != self._config.device:
                    self.device = device
                return

            if device == self._config.device and signal_text == self._config.signal:
                return

            # Get the signal config stored in the combobox
            signal_config = device_selection.signal_combo_box.get_signal_config()

            if not signal_config:
                # Fallback: try to get config from device
                try:
                    device_obj = self.dev[device]
                    signal_config = device_obj._info["signals"].get(signal_text, {})
                except (KeyError, AttributeError):
                    logger.warning(f"Could not get signal config for {device}.{signal_text}")
                    signal_config = None

            # Store signal config and set properties which will trigger the connection
            self._signal_configs["main"] = signal_config
            self.device = device
            self.signal = signal_text
        finally:
            self._device_selection_updating = False

    ################################################################################
    # Data Acquisition

    @SafeProperty(str, auto_emit=True)
    def device(self) -> str:
        """
        The name of the device to monitor for image data.
        """
        return self._config.device

    @device.setter
    def device(self, value: str):
        """
        Set the device name for the image. This should be used together with signal.
        When both device and signal are set, the widget connects to that device signal.

        Args:
            value(str): The name of the device to monitor.
        """
        if not value:
            # Clear the monitor if empty device name
            if self._config.device:
                self._disconnect_current_monitor()
                self._config.device = ""
                self._config.signal = ""
                self._signal_configs.pop("main", None)
                self._set_connection_status("disconnected")
            return

        old_device = self._config.device
        old_signal = self._config.signal
        old_config = self.subscriptions["main"]
        if old_device and old_signal and old_device != value:
            self._disconnect_monitor_subscription(
                device=old_device,
                signal=old_signal,
                source=old_config.source,
                async_update=self.async_update,
                async_signal_name=old_config.async_signal_name,
            )
        self._config.device = value

        # If we have a signal, reconnect with the new device
        if self._config.signal:
            # Try to get fresh signal config for the new device
            try:
                device_obj = self.dev[value]
                # Try to get signal config for the current entry
                if self._config.signal in device_obj._info.get("signals", {}):
                    self._signal_configs["main"] = device_obj._info["signals"][self._config.signal]
                    self._setup_connection()
                else:
                    # Signal doesn't exist on new device
                    logger.warning(
                        f"Signal '{self._config.signal}' doesn't exist on device '{value}'"
                    )
                    self._disconnect_current_monitor()
                    self._config.signal = ""
                    self._signal_configs.pop("main", None)
                    self._set_connection_status(
                        "error", f"Signal '{self._config.signal}' doesn't exist"
                    )
            except (KeyError, AttributeError):
                # Device doesn't exist
                logger.warning(f"Device '{value}' not found")
                if old_device:
                    self._disconnect_current_monitor()
                self._set_connection_status("error", f"Device '{value}' not found")

        # Toolbar sync happens via SafeProperty auto_emit property_changed handling.

    @SafeProperty(str, auto_emit=True)
    def signal(self) -> str:
        """
        The signal/entry name to monitor on the device.
        """
        return self._config.signal

    @signal.setter
    def signal(self, value: str):
        """
        Set the device signal for the image. This should be used together with device.
        When set, it will connect to updates from that device signal.

        Args:
            value(str): The signal name to monitor.
        """
        if not value:
            if self._config.signal:
                self._disconnect_current_monitor()
                self._config.signal = ""
                self._signal_configs.pop("main", None)
                self._set_connection_status("disconnected")
            return

        old_signal = self._config.signal
        old_config = self.subscriptions["main"]
        if self._config.device and old_signal and old_signal != value:
            self._disconnect_monitor_subscription(
                device=self._config.device,
                signal=old_signal,
                source=old_config.source,
                async_update=self.async_update,
                async_signal_name=old_config.async_signal_name,
            )
        self._config.signal = value

        # If we have a device, try to connect
        if self._config.device:
            try:
                device_obj = self.dev[self._config.device]
                signal_config = device_obj._info["signals"].get(value)
                if not isinstance(signal_config, dict) or not signal_config.get("signal_class"):
                    logger.warning(
                        f"Could not find valid configuration for signal '{value}' "
                        f"on device '{self._config.device}'."
                    )
                    self._signal_configs.pop("main", None)
                    self._set_connection_status("error", f"Signal '{value}' not found")
                    return

                self._signal_configs["main"] = signal_config
                self._setup_connection()
            except (KeyError, AttributeError):
                logger.warning(
                    f"Could not find signal '{value}' on device '{self._config.device}'."
                )
                # Remove signal config if it can't be fetched
                self._signal_configs.pop("main", None)
                self._set_connection_status("error", f"Signal '{value}' not found")

        else:
            logger.debug(f"signal setter: No device set yet for signal '{value}'")

    @property
    def main_image(self) -> ImageItem:
        """Access the main image item."""
        return self.layer_manager["main"].image

    def _setup_connection(self):
        """
        Internal method to setup connection based on current device, signal, and signal_config.
        """
        if not self._config.device or not self._config.signal:
            logger.warning("Cannot setup connection without both device and signal")
            self._set_connection_status("disconnected")
            return

        signal_config = self._signal_configs.get("main")
        if not signal_config:
            logger.warning(
                f"Cannot setup connection for {self._config.device}.{self._config.signal} without signal_config"
            )
            self._set_connection_status("error", "Missing signal config")
            return

        # Disconnect any existing monitor first
        self._disconnect_current_monitor()

        # Determine monitor type and source from signal_config
        signal_class = signal_config.get("signal_class", None)
        supported_classes = ["PreviewSignal"] + self.SUPPORTED_SIGNALS

        if signal_class not in supported_classes:
            logger.warning(
                f"Signal '{self._config.device}.{self._config.signal}' has unsupported signal class '{signal_class}'. "
                f"Supported classes: {supported_classes}"
            )
            self._set_connection_status("error", f"Unsupported signal class '{signal_class}'")
            return

        describe = signal_config.get("describe") or {}
        signal_info = describe.get("signal_info") or {}
        ndim = signal_info.get("ndim", None)

        if ndim is None:
            logger.warning(
                f"Signal '{self._config.device}.{self._config.signal}' does not have a valid 'ndim' in its signal_info."
            )
            self._set_connection_status("error", "Missing ndim in signal_info")
            return

        config = self.subscriptions["main"]
        self.async_update = False
        config.async_signal_name = None

        if ndim == 1:
            config.source = "device_monitor_1d"
            config.monitor_type = "1d"
            if signal_class == "PreviewSignal":
                self.bec_dispatcher.connect_slot(
                    self.on_image_update_1d,
                    MessageEndpoints.device_preview(self._config.device, self._config.signal),
                )
            elif signal_class in self.SUPPORTED_SIGNALS:
                self.async_update = True
                config.async_signal_name = signal_config.get(
                    "obj_name", f"{self._config.device}_{self._config.signal}"
                )
                self._setup_async_image(self.scan_id)
        elif ndim == 2:
            config.source = "device_monitor_2d"
            config.monitor_type = "2d"
            if signal_class == "PreviewSignal":
                self.bec_dispatcher.connect_slot(
                    self.on_image_update_2d,
                    MessageEndpoints.device_preview(self._config.device, self._config.signal),
                )
            elif signal_class in self.SUPPORTED_SIGNALS:
                self.async_update = True
                config.async_signal_name = signal_config.get(
                    "obj_name", f"{self._config.device}_{self._config.signal}"
                )
                self._setup_async_image(self.scan_id)
        else:
            logger.warning(
                f"Unsupported ndim '{ndim}' for monitor '{self._config.device}.{self._config.signal}'."
            )
            self._set_connection_status("error", f"Unsupported ndim '{ndim}'")
            return

        self._set_connection_status("connected")
        logger.info(
            f"Connected to {self._config.device}.{self._config.signal} with type {config.monitor_type}"
        )
        self._autorange_on_next_update = True

    def _disconnect_monitor_subscription(
        self,
        *,
        device: str,
        signal: str,
        source: Literal["device_monitor_1d", "device_monitor_2d"] | None,
        async_update: bool,
        async_signal_name: str | None,
    ) -> None:
        if not device or not signal:
            return

        if async_update:
            async_signal_name = async_signal_name or signal
            ids_to_check = [self.scan_id, self.old_scan_id]

            if source == "device_monitor_1d":
                for scan_id in ids_to_check:
                    if scan_id is None:
                        continue
                    self.bec_dispatcher.disconnect_slot(
                        self.on_image_update_1d,
                        MessageEndpoints.device_async_signal(scan_id, device, async_signal_name),
                    )
                    logger.info(
                        f"Disconnecting 1d update ScanID:{scan_id}, Device Name:{device},Device Entry:{async_signal_name}"
                    )
            elif source == "device_monitor_2d":
                for scan_id in ids_to_check:
                    if scan_id is None:
                        continue
                    self.bec_dispatcher.disconnect_slot(
                        self.on_image_update_2d,
                        MessageEndpoints.device_async_signal(scan_id, device, async_signal_name),
                    )
                    logger.info(
                        f"Disconnecting 2d update ScanID:{scan_id}, Device Name:{device},Device Entry:{async_signal_name}"
                    )
            return

        if source == "device_monitor_1d":
            self.bec_dispatcher.disconnect_slot(
                self.on_image_update_1d, MessageEndpoints.device_preview(device, signal)
            )
            logger.info(
                f"Disconnecting preview 1d update Device Name:{device}, Device Entry:{signal}"
            )
        elif source == "device_monitor_2d":
            self.bec_dispatcher.disconnect_slot(
                self.on_image_update_2d, MessageEndpoints.device_preview(device, signal)
            )
            logger.info(
                f"Disconnecting preview 2d update Device Name:{device}, Device Entry:{signal}"
            )

    def _disconnect_current_monitor(self):
        """
        Internal method to disconnect the current monitor subscriptions.
        """
        if not self._config.device or not self._config.signal:
            return

        config = self.subscriptions["main"]
        self._disconnect_monitor_subscription(
            device=self._config.device,
            signal=self._config.signal,
            source=config.source,
            async_update=self.async_update,
            async_signal_name=config.async_signal_name,
        )

        # Reset async state
        self.async_update = False
        config.async_signal_name = None
        self._set_connection_status("disconnected")

    ################################################################################
    # High Level methods for API
    ################################################################################
    @SafeSlot(popup_error=True)
    def image(
        self,
        device: str | None = None,
        signal: str | None = None,
        color_map: str | None = None,
        color_bar: Literal["simple", "full"] | None = None,
        vrange: tuple[int, int] | None = None,
    ) -> ImageItem | None:
        """
        Set the image source and update the image.

        Args:
            device(str|None): The name of the device to monitor. If None or empty string, the current monitor will be disconnected.
            signal(str|None): The signal/entry name to monitor on the device.
            color_map(str): The color map to use for the image.
            color_bar(str): The type of color bar to use. Options are "simple" or "full".
            vrange(tuple): The range of values to use for the color map.

        Returns:
            ImageItem: The image object, or None if connection failed.
        """
        # Disconnect existing monitor if any
        if self._config.device and self._config.signal:
            self._disconnect_current_monitor()

        if not device or not signal:
            if device or signal:
                logger.warning("Both device and signal must be specified")
            else:
                logger.info("Disconnecting image monitor")
            self.device = ""
            return None

        # Validate device
        self.entry_validator.validate_monitor(device)

        # Clear old entry first to avoid reconnect attempts on the new device
        if self._config.signal:
            self.signal = ""

        # Set properties to trigger connection
        self.device = device
        self.signal = signal

        # Apply visual settings
        if color_map is not None:
            self.main_image.color_map = color_map
        if color_bar is not None:
            self.enable_colorbar(True, color_bar)
        if vrange is not None:
            self.vrange = vrange

        return self.main_image

    def _sync_device_selection(self):
        """
        Synchronize the device and signal comboboxes with the current monitor state.
        This ensures the toolbar reflects the device and signal properties.
        """
        try:
            device_selection_action = self.toolbar.components.get_action("device_selection")
        except Exception:  # noqa: BLE001 - toolbar might not be ready during early init
            logger.warning(f"Image ({self.object_name}) toolbar was not ready during init.")
            return

        if device_selection_action is None:
            return

        device_selection: DeviceSelection = device_selection_action.widget
        target_device = self._config.device or ""
        target_entry = self._config.signal or ""

        # Check if already synced
        if (
            device_selection.device_combo_box.currentText() == target_device
            and device_selection.signal_combo_box.currentText() == target_entry
        ):
            return

        device_selection.set_device_and_signal(target_device, target_entry)

    def _sync_signal_from_toolbar(self) -> None:
        """
        Pull the signal selection from the toolbar if it differs from the current signal.
        This keeps CLI-driven device updates in sync with the signal combobox state.
        """
        if self._device_selection_updating:
            return

        if not self._config.device:
            return

        try:
            device_selection_action = self.toolbar.components.get_action("device_selection")
        except Exception:  # noqa: BLE001 - toolbar might not be ready during early init
            return

        if device_selection_action is None:
            return

        device_selection: DeviceSelection = device_selection_action.widget
        if device_selection.device_combo_box.currentText() != self._config.device:
            return

        signal_text = device_selection.signal_combo_box.currentText()
        if not signal_text or signal_text == self._config.signal:
            return

        signal_config = device_selection.signal_combo_box.get_signal_config()
        if not signal_config:
            try:
                device_obj = self.dev[self._config.device]
                signal_config = device_obj._info["signals"].get(signal_text, {})
            except (KeyError, AttributeError):
                signal_config = None

        if not signal_config:
            return

        self._signal_configs["main"] = signal_config
        self._device_selection_updating = True
        try:
            self.signal = signal_text
        finally:
            self._device_selection_updating = False

    def _set_connection_status(self, status: str, message: str | None = None) -> None:
        self._config.connection_status = status
        self._config.connection_error = message
        self.property_changed.emit("connection_status", status)
        self.property_changed.emit("connection_error", message or "")

    ################################################################################
    # Post Processing
    ################################################################################

    @SafeProperty(bool, auto_emit=True)
    def fft(self) -> bool:
        """
        Whether FFT postprocessing is enabled.
        """
        return self.main_image.fft

    @fft.setter
    def fft(self, enable: bool):
        """
        Set FFT postprocessing.

        Args:
            enable(bool): Whether to enable FFT postprocessing.
        """
        self.main_image.fft = enable

    @SafeProperty(bool, auto_emit=True)
    def log(self) -> bool:
        """
        Whether logarithmic scaling is applied.
        """
        return self.main_image.log

    @log.setter
    def log(self, enable: bool):
        """
        Set logarithmic scaling.

        Args:
            enable(bool): Whether to enable logarithmic scaling.
        """
        self.main_image.log = enable

    @SafeProperty(int)
    def num_rotation_90(self) -> int:
        """
        The number of 90° rotations to apply counterclockwise.
        """
        return self.main_image.num_rotation_90

    @num_rotation_90.setter
    def num_rotation_90(self, value: int):
        """
        Set the number of 90° rotations to apply counterclockwise.

        Args:
            value(int): The number of 90° rotations to apply.
        """
        self.main_image.num_rotation_90 = value

    @SafeProperty(bool, auto_emit=True)
    def transpose(self) -> bool:
        """
        Whether the image is transposed.
        """
        return self.main_image.transpose

    @transpose.setter
    def transpose(self, enable: bool):
        """
        Set the image to be transposed.

        Args:
            enable(bool): Whether to enable transposing the image.
        """
        self.main_image.transpose = enable

    ################################################################################
    # Image Update Methods
    ################################################################################

    ########################################
    # Connections

    @SafeSlot(dict, dict)
    def on_scan_status(self, msg: dict, meta: dict):
        """
        Initial scan status message handler, which is triggered at the beginning and end of scan.
        Needed for setup of AsyncSignal connections.

        Args:
            msg(dict): The message content.
            meta(dict): The message metadata.
        """
        current_scan_id = msg.get("scan_id", None)
        if current_scan_id is None:
            return
        self._handle_scan_change(current_scan_id)

    @SafeSlot(dict, dict)
    def on_scan_progress(self, msg: dict, meta: dict):
        """
        For setting async image readback during scan progress updates if widget is started later than scan.

        Args:
            msg(dict): The message content.
            meta(dict): The message metadata.
        """
        current_scan_id = meta.get("scan_id", None)
        if current_scan_id is None:
            return
        self._handle_scan_change(current_scan_id)

    def _handle_scan_change(self, current_scan_id: str):
        """
        Update internal scan ids and refresh async connections if needed.
        Also clears image buffers when scan changes.

        Args:
            current_scan_id (str): The current scan identifier.
        """
        if current_scan_id == self.scan_id:
            return

        # Scan ID changed - clear buffers and reset image
        self.old_scan_id = self.scan_id
        self.scan_id = current_scan_id

        # Clear image buffer for 1D data accumulation
        self.main_image.clear()
        if hasattr(self.main_image, "buffer"):
            self.main_image.buffer = []
            self.main_image.max_len = 0

        # Reset crosshair if present
        if self.crosshair is not None:
            self.crosshair.reset()

        # Reconnect async image subscription with new scan_id
        if self.async_update:
            self._setup_async_image(scan_id=self.scan_id)

    def _get_async_signal_name(self) -> tuple[str, str] | None:
        """
        Returns device and async signal names used for endpoints/messages.

        Returns:
            tuple[str, str] | None: (device, async_signal_name) or None if not available.
        """
        if not self._config.device or not self._config.signal:
            return None

        config = self.subscriptions["main"]
        async_signal = config.async_signal_name or self._config.signal
        return self._config.device, async_signal

    def _setup_async_image(self, scan_id: str | None):
        """
        (Re)connect async image readback for the current scan.

        Args:
            scan_id (str | None): The scan identifier to subscribe to.
        """
        if not self.async_update:
            return

        config = self.subscriptions["main"]
        async_names = self._get_async_signal_name()
        if async_names is None:
            logger.info("Async image setup skipped because monitor information is incomplete.")
            return

        device, async_signal = async_names
        if config.monitor_type == "1d":
            slot = self.on_image_update_1d
        elif config.monitor_type == "2d":
            slot = self.on_image_update_2d
        else:
            logger.warning(
                f"Async image setup skipped due to unsupported monitor type '{config.monitor_type}'."
            )
            return

        # Disconnect any previous scan subscriptions to avoid stale updates.
        for prev_scan_id in (self.old_scan_id, self.scan_id):
            if prev_scan_id is None:
                continue
            self.bec_dispatcher.disconnect_slot(
                slot, MessageEndpoints.device_async_signal(prev_scan_id, device, async_signal)
            )

        if scan_id is None:
            logger.info("Scan ID not available yet; delaying async image subscription.")
            return

        self.bec_dispatcher.connect_slot(
            slot,
            MessageEndpoints.device_async_signal(scan_id, device, async_signal),
            from_start=True,
            cb_info={"scan_id": scan_id},
        )
        logger.info(f"Setup async image for {device}.{async_signal} and scan {scan_id}.")

    def disconnect_monitor(self, device: str | None = None, signal: str | None = None):
        """
        Disconnect the monitor from the image update signals, both 1D and 2D.

        Args:
            device(str|None): The name of the device to disconnect. Defaults to current device.
            signal(str|None): The signal/entry name to disconnect. Defaults to current signal.
        """
        config = self.subscriptions["main"]
        target_device = device or self._config.device
        target_entry = signal or self._config.signal

        if not target_device or not target_entry:
            logger.warning("Cannot disconnect monitor without both device and signal")
            return

        if config.source not in {"device_monitor_1d", "device_monitor_2d"}:
            logger.warning(
                f"Cannot disconnect monitor {target_device}.{target_entry} with source {self.subscriptions['main'].source}"
            )
            return

        self._disconnect_monitor_subscription(
            device=target_device,
            signal=target_entry,
            source=config.source,
            async_update=self.async_update,
            async_signal_name=config.async_signal_name,
        )

        self.subscriptions["main"].async_signal_name = None
        self.async_update = False
        self._sync_device_selection()

    ########################################
    # 1D updates

    @SafeSlot(dict, dict)
    def on_image_update_1d(self, msg: dict, metadata: dict):
        """
        Update the image with 1D data.
        For preview signals: metadata doesn't contain scan_id.
        For async signals: scan_id is managed via on_scan_status/on_scan_progress.

        Args:
            msg(dict): The message containing the data.
            metadata(dict): The metadata associated with the message.
        """
        try:
            image = self.main_image
        except Exception:
            return
        data = self._get_payload_data(msg)

        if data is None:
            logger.warning("No data received for image update from 1D.")
            return

        image_buffer = self.adjust_image_buffer(image, data)

        if self._color_bar is not None:
            self._color_bar.blockSignals(True)
        image.set_data(image_buffer)
        if self._color_bar is not None:
            self._color_bar.blockSignals(False)
        if self._autorange_on_next_update:
            self._autorange_on_next_update = False
            self.auto_range()
        self.image_updated.emit()

    @staticmethod
    def adjust_image_buffer(image: ImageItem, new_data: np.ndarray) -> np.ndarray:
        """
        Adjusts the image buffer to accommodate the new data, ensuring that all rows have the same length.

        Args:
            image: The image object (used to store a buffer list and max_len).
            new_data (np.ndarray): The new incoming 1D waveform data.

        Returns:
            np.ndarray: The updated image buffer with adjusted shapes.
        """
        new_len = new_data.shape[0]
        if not hasattr(image, "buffer"):
            image.buffer = []
            image.max_len = 0

        if new_len > image.max_len:
            image.max_len = new_len
            for i in range(len(image.buffer)):
                wf = image.buffer[i]
                pad_width = image.max_len - wf.shape[0]
                if pad_width > 0:
                    image.buffer[i] = np.pad(wf, (0, pad_width), mode="constant", constant_values=0)
            image.buffer.append(new_data)
        else:
            pad_width = image.max_len - new_len
            if pad_width > 0:
                new_data = np.pad(new_data, (0, pad_width), mode="constant", constant_values=0)
            image.buffer.append(new_data)

        image_buffer = np.array(image.buffer)
        return image_buffer

    ########################################
    # 2D updates

    @SafeSlot(dict, dict)
    def on_image_update_2d(self, msg: dict, metadata: dict):
        """
        Update the image with 2D data.

        Args:
            msg(dict): The message containing the data.
            metadata(dict): The metadata associated with the message.
        """
        try:
            image = self.main_image
        except Exception:
            return
        data = self._get_payload_data(msg)
        if data is None:
            logger.warning("No data received for image update from 2D.")
            return
        if self._color_bar is not None:
            self._color_bar.blockSignals(True)
        image.set_data(data)
        if self._color_bar is not None:
            self._color_bar.blockSignals(False)
        if self._autorange_on_next_update:
            self._autorange_on_next_update = False
            self.auto_range()
        self.image_updated.emit()

    def _get_payload_data(self, msg: dict) -> np.ndarray | None:
        """
        Extract payload from async/preview/monitor1D/2D message structures due to inconsistent formats in backend.

        Args:
            msg (dict): The incoming message containing data.
        """
        if not self.async_update:
            return msg.get("data")
        async_names = self._get_async_signal_name()
        if async_names is None:
            logger.warning("Async payload extraction failed; monitor info incomplete.")
            return None
        _, async_signal = async_names
        return msg.get("signals", {}).get(async_signal, {}).get("value", None)

    ################################################################################
    # Clean up
    ################################################################################

    @SafeSlot(str)
    def _on_layer_removed(self, layer_name: str):
        """
        Handle the removal of a layer by disconnecting the monitor.

        Args:
            layer_name(str): The name of the layer that was removed.
        """
        if layer_name not in self.subscriptions:
            return
        # For the main layer, disconnect current monitor
        if layer_name == "main" and self._config.device and self._config.signal:
            self._disconnect_current_monitor()
            self._config.device = ""
            self._config.signal = ""
            self._signal_configs.pop("main", None)

    def cleanup(self):
        """
        Disconnect the image update signals and clean up the image.
        """
        self.layer_removed.disconnect(self._on_layer_removed)

        # Disconnect current monitor
        if self._config.device and self._config.signal:
            self._disconnect_current_monitor()

        self.subscriptions.clear()

        # Toolbar cleanup - disconnect the device_selection bundle
        try:
            self.toolbar.disconnect_bundle("device_selection")
        except Exception:  # noqa: BLE001
            pass

        self.bec_dispatcher.disconnect_slot(self.on_scan_status, MessageEndpoints.scan_status())
        self.bec_dispatcher.disconnect_slot(self.on_scan_progress, MessageEndpoints.scan_progress())
        super().cleanup()


if __name__ == "__main__":  # pragma: no cover
    import sys

    from qtpy.QtWidgets import QApplication, QHBoxLayout

    app = QApplication(sys.argv)
    apply_theme("dark")
    win = QWidget()
    win.setWindowTitle("Image Demo")
    ml = QHBoxLayout(win)

    image_popup = Image(popups=True)
    # image_side_panel = Image(popups=False)

    ml.addWidget(image_popup)
    # ml.addWidget(image_side_panel)

    win.resize(1500, 800)
    win.show()
    sys.exit(app.exec_())



from bec_widgets.widgets.editors.website.website import WebsiteWidget


class SBBMonitor(WebsiteWidget):
    """
    A widget to display the SBB monitor website.
    """

    PLUGIN = True
    ICON_NAME = "train"
    USER_ACCESS = []

    def __init__(self, parent=None, **kwargs):
        url = "https://free.oevplus.ch/monitor/?viewType=splitView&layout=1&showClock=true&showPerron=true&stationGroup1Title=Villigen%2C%20PSI%20West&stationGroup2Title=Siggenthal-Würenlingen&station_1_id=85%3A3592&station_1_name=Villigen%2C%20PSI%20West&station_1_quantity=5&station_1_group=1&station_2_id=85%3A3502&station_2_name=Siggenthal-Würenlingen&station_2_quantity=5&station_2_group=2"
        super().__init__(parent=parent, url=url, **kwargs)



from bec_widgets.widgets.control.device_control.positioner_box.positioner_box.positioner_box import (
    PositionerBox,
)
from bec_widgets.widgets.control.device_control.positioner_box.positioner_box_2d.positioner_box_2d import (
    PositionerBox2D,
)
from bec_widgets.widgets.control.device_control.positioner_box.positioner_control_line.positioner_control_line import (
    PositionerControlLine,
)

__ALL__ = ["PositionerBox", "PositionerControlLine", "PositionerBox2D"]



from __future__ import annotations

from typing import TYPE_CHECKING

import pyqtgraph as pg

from bec_widgets.utils.error_popups import SafeSlot
from bec_widgets.utils.toolbars.actions import MaterialIconAction, SwitchableToolBarAction
from bec_widgets.utils.toolbars.bundles import ToolbarBundle
from bec_widgets.utils.toolbars.connections import BundleConnection

if TYPE_CHECKING:
    from bec_widgets.utils.toolbars.toolbar import ToolbarComponents


def mouse_interaction_bundle(components: ToolbarComponents) -> ToolbarBundle:
    """
    Creates a mouse interaction toolbar bundle.

    Args:
        components (ToolbarComponents): The components to be added to the bundle.

    Returns:
        ToolbarBundle: The mouse interaction toolbar bundle.
    """
    components.add_safe(
        "mouse_drag",
        MaterialIconAction(
            icon_name="drag_pan",
            tooltip="Drag Mouse Mode",
            checkable=True,
            parent=components.toolbar,
        ),
    )
    components.add_safe(
        "mouse_rect",
        MaterialIconAction(
            icon_name="frame_inspect",
            tooltip="Rectangle Zoom Mode",
            checkable=True,
            parent=components.toolbar,
        ),
    )
    components.add_safe(
        "auto_range",
        MaterialIconAction(
            icon_name="open_in_full",
            tooltip="Autorange Plot",
            checkable=False,
            parent=components.toolbar,
        ),
    )
    components.add_safe(
        "switch_mouse_mode",
        SwitchableToolBarAction(
            actions={
                "drag_mode": components.get_action_reference("mouse_drag")(),
                "rectangle_mode": components.get_action_reference("mouse_rect")(),
            },
            initial_action="drag_mode",
            tooltip="Mouse Modes",
            checkable=True,
            parent=components.toolbar,
            default_state_checked=True,
        ),
    )
    bundle = ToolbarBundle("mouse_interaction", components)
    bundle.add_action("switch_mouse_mode")
    bundle.add_action("auto_range")
    return bundle


class MouseInteractionConnection(BundleConnection):
    """
    Connection class for mouse interaction toolbar bundle.
    """

    def __init__(self, components: ToolbarComponents, target_widget=None):
        self.bundle_name = "mouse_interaction"
        self.components = components
        self.target_widget = target_widget
        self.mouse_mode = None
        if (
            not hasattr(self.target_widget, "plot_item")
            or not hasattr(self.target_widget, "auto_range_x")
            or not hasattr(self.target_widget, "auto_range_y")
        ):
            raise AttributeError(
                "Target widget must implement required methods for mouse interactions."
            )
        super().__init__()
        self._connected = False  # Track if the connection has been made

    def connect(self):
        self._connected = True
        drag = self.components.get_action_reference("mouse_drag")()
        rect = self.components.get_action_reference("mouse_rect")()
        auto = self.components.get_action_reference("auto_range")()

        drag.action.toggled.connect(self.enable_mouse_pan_mode)
        rect.action.toggled.connect(self.enable_mouse_rectangle_mode)
        auto.action.triggered.connect(self.autorange_plot)

    def disconnect(self):
        if not self._connected:
            return
        # Disconnect the action from the target widget's method
        drag = self.components.get_action_reference("mouse_drag")()
        rect = self.components.get_action_reference("mouse_rect")()
        auto = self.components.get_action_reference("auto_range")()
        drag.action.toggled.disconnect(self.enable_mouse_pan_mode)
        rect.action.toggled.disconnect(self.enable_mouse_rectangle_mode)
        auto.action.triggered.disconnect(self.autorange_plot)

    def get_viewbox_mode(self):
        """
        Returns the current interaction mode of a PyQtGraph ViewBox and sets the corresponding action.
        """

        if self.target_widget:
            viewbox = self.target_widget.plot_item.getViewBox()
            switch_mouse_action = self.components.get_action_reference("switch_mouse_mode")()
            if viewbox.getState()["mouseMode"] == 3:
                switch_mouse_action.set_default_action("drag_mode")
                switch_mouse_action.main_button.setChecked(True)
                self.mouse_mode = "PanMode"
            elif viewbox.getState()["mouseMode"] == 1:
                switch_mouse_action.set_default_action("rectangle_mode")
                switch_mouse_action.main_button.setChecked(True)
                self.mouse_mode = "RectMode"

    @SafeSlot(bool)
    def enable_mouse_rectangle_mode(self, checked: bool):
        """
        Enable the rectangle zoom mode on the plot widget.
        """
        switch_mouse_action = self.components.get_action_reference("switch_mouse_mode")()
        if self.mouse_mode == "RectMode":
            switch_mouse_action.main_button.setChecked(True)
            return
        drag_mode = self.components.get_action_reference("mouse_drag")()
        drag_mode.action.setChecked(not checked)
        if self.target_widget and checked:
            self.target_widget.plot_item.getViewBox().setMouseMode(pg.ViewBox.RectMode)
            self.mouse_mode = "RectMode"

    @SafeSlot(bool)
    def enable_mouse_pan_mode(self, checked: bool):
        """
        Enable the pan mode on the plot widget.
        """
        if self.mouse_mode == "PanMode":
            switch_mouse_action = self.components.get_action_reference("switch_mouse_mode")()
            switch_mouse_action.main_button.setChecked(True)
            return
        rect_mode = self.components.get_action_reference("mouse_rect")()
        rect_mode.action.setChecked(not checked)
        if self.target_widget and checked:
            self.target_widget.plot_item.getViewBox().setMouseMode(pg.ViewBox.PanMode)
            self.mouse_mode = "PanMode"

    @SafeSlot()
    def autorange_plot(self):
        """
        Enable autorange on the plot widget.
        """
        if self.target_widget:
            self.target_widget.auto_range()



from __future__ import annotations

from types import GenericAlias, NoneType, UnionType
from typing import NamedTuple

from bec_lib.logger import bec_logger
from bec_qthemes import material_icon
from pydantic import BaseModel, ValidationError
from qtpy.QtCore import Signal  # type: ignore
from qtpy.QtWidgets import QApplication, QGridLayout, QLabel, QSizePolicy, QVBoxLayout, QWidget

from bec_widgets.utils.bec_widget import BECWidget
from bec_widgets.utils.compact_popup import CompactPopupWidget
from bec_widgets.utils.error_popups import SafeProperty, SafeSlot
from bec_widgets.utils.forms_from_types import styles
from bec_widgets.utils.forms_from_types.items import (
    DynamicFormItem,
    DynamicFormItemType,
    FormItemSpec,
    widget_from_type,
)

logger = bec_logger.logger


class GridRow(NamedTuple):
    i: int
    label: QLabel
    widget: DynamicFormItem


class TypedForm(BECWidget, QWidget):
    PLUGIN = True
    ICON_NAME = "list_alt"

    value_changed = Signal()

    RPC = True
    USER_ACCESS = ["enabled", "enabled.setter"]

    def __init__(
        self,
        parent=None,
        items: list[tuple[str, type]] | None = None,
        form_item_specs: list[FormItemSpec] | None = None,
        enabled: bool = True,
        pretty_display: bool = False,
        client=None,
        **kwargs,
    ):
        """Widget with a list of form items based on a list of types.

        Args:
            items (list[tuple[str, type]]):         list of tuples of a name for the field and its type.
                                                    Should be a type supported by the logic in items.py
            form_item_specs (list[FormItemSpec]):   list of form item specs, equivalent to items.
                                                    only one of items or form_item_specs should be
                                                    supplied.
            enabled (bool, optional):               whether fields are enabled for editing.
            pretty_display (bool, optional): Whether to use a pretty display for the widget. Defaults to False. If True, disables the widget, doesn't add a clear button, and adapts the stylesheet for non-editable display.
        """
        if items is not None and form_item_specs is not None:
            logger.error(
                "Must specify one and only one of items and form_item_specs! Ignoring `items`."
            )
            items = None
        if items is None and form_item_specs is None:
            logger.error("Must specify one and only one of items and form_item_specs!")
            items = []
        super().__init__(parent=parent, client=client, **kwargs)
        self._items = form_item_specs or [
            FormItemSpec(name=name, item_type=item_type, pretty_display=pretty_display)
            for name, item_type in items  # type: ignore
        ]
        self.setSizePolicy(QSizePolicy.MinimumExpanding, QSizePolicy.Minimum)
        self._layout = QVBoxLayout()
        self._layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(self._layout)

        self._enabled: bool = enabled

        self._form_grid_container = QWidget(parent=self)
        self._form_grid_container.setSizePolicy(QSizePolicy.MinimumExpanding, QSizePolicy.Minimum)
        self._form_grid_container.setLayout(QVBoxLayout())
        self._layout.addWidget(self._form_grid_container)

        self._form_grid = QWidget(parent=self._form_grid_container)
        self._form_grid.setSizePolicy(QSizePolicy.MinimumExpanding, QSizePolicy.Minimum)
        self._form_grid.setLayout(self._new_grid_layout())

        self._widget_types: dict | None = None
        self._widget_from_type = widget_from_type
        self._post_init()

    def _post_init(self):
        """Override this if a subclass should do things after super().__init__ and before populate()"""
        self.populate()
        self.enabled = self._enabled  # type: ignore # QProperty

    def populate(self):
        self._clear_grid()
        for r, item in enumerate(self._items):
            self._add_griditem(item, r)
        gl: QGridLayout = self._form_grid.layout()
        gl.setRowStretch(gl.rowCount(), 1)

    def _add_griditem(self, item: FormItemSpec, row: int):
        grid = self._form_grid.layout()
        # Use title from FieldInfo if available, otherwise use the property name
        label_text = item.info.title if item.info.title else item.name
        label = QLabel(parent=self._form_grid, text=label_text)
        label.setProperty("_model_field_name", item.name)
        label.setToolTip(item.info.description or item.name)
        grid.addWidget(label, row, 0)
        widget = self._widget_from_type(item, self._widget_types)(parent=self._form_grid, spec=item)
        widget.valueChanged.connect(self.value_changed)
        widget.setSizePolicy(QSizePolicy.MinimumExpanding, QSizePolicy.Minimum)
        grid.addWidget(widget, row, 1)

    def enumerate_form_widgets(self):
        """Return a generator over the rows of the form, with the row number, the label widget (to
        which the field name is attached as a property "_model_field_name"), and the entry widget"""
        grid: QGridLayout = self._form_grid.layout()  # type: ignore
        for i in range(grid.rowCount() - 1):  # One extra row for stretch
            yield GridRow(i, grid.itemAtPosition(i, 0).widget(), grid.itemAtPosition(i, 1).widget())

    def _dict_from_grid(self) -> dict[str, DynamicFormItemType]:
        return {
            row.label.property("_model_field_name"): row.widget.getValue()
            for row in self.enumerate_form_widgets()
        }

    def _clear_grid(self):
        gl = self._form_grid.layout()
        while w := gl.takeAt(0):
            w = w.widget()
            if hasattr(w, "teardown"):
                w.teardown()
            w.deleteLater()
        self._form_grid_container.layout().removeWidget(self._form_grid)
        self._form_grid.deleteLater()
        self._form_grid = QWidget()
        self._form_grid.setLayout(self._new_grid_layout())
        self._form_grid_container.layout().addWidget(self._form_grid)
        self.update_size()

    def update_size(self):
        self._form_grid.adjustSize()
        self._form_grid_container.adjustSize()
        self.adjustSize()

    def _new_grid_layout(self):
        new_grid = QGridLayout(self)
        new_grid.setContentsMargins(0, 0, 0, 0)
        return new_grid

    @property
    def widget_dict(self):
        return {
            row.label.property("_model_field_name"): row.widget
            for row in self.enumerate_form_widgets()
        }

    @SafeProperty(bool)
    def enabled(self):
        return self._enabled

    @enabled.setter
    def enabled(self, value: bool):
        self._enabled = value
        self.setEnabled(value)


class PydanticModelForm(TypedForm):
    form_data_updated = Signal(dict)
    form_data_cleared = Signal(NoneType)
    validity_proc = Signal(bool)

    def __init__(
        self,
        parent=None,
        data_model: type[BaseModel] | None = None,
        enabled: bool = True,
        pretty_display: bool = False,
        client=None,
        **kwargs,
    ):
        """
        A form generated from a pydantic model.

        Args:
            data_model (type[BaseModel]): the model class for which to generate a form.
            enabled (bool, optional): whether fields are enabled for editing.
            pretty_display (bool, optional): Whether to use a pretty display for the widget. Defaults to False. If True, disables the widget, doesn't add a clear button, and adapts the stylesheet for non-editable display.

        """
        self._pretty_display = pretty_display
        self._md_schema = data_model
        super().__init__(
            parent=parent,
            form_item_specs=self._form_item_specs(),
            enabled=enabled,
            client=client,
            **kwargs,
        )

        self._validity = CompactPopupWidget()
        self._validity.compact_view = True  # type: ignore
        self._validity.label = "Validity"  # type: ignore
        self._validity.compact_show_popup.setIcon(
            material_icon(icon_name="info", size=(10, 10), convert_to_pixmap=False)
        )
        self._validity_message = QLabel("Not yet validated")
        self._validity.addWidget(self._validity_message)
        self._layout.addWidget(self._validity)
        self.value_changed.connect(self.validate_form)

        self._connect_to_theme_change()

    @SafeSlot()
    def clear(self): ...

    def set_pretty_display_theme(self, theme: str = "dark"):
        if self._pretty_display:
            self.setStyleSheet(styles.pretty_display_theme(theme))

    def _connect_to_theme_change(self):
        """Connect to the theme change signal."""
        qapp = QApplication.instance()
        if hasattr(qapp, "theme_signal"):
            qapp.theme_signal.theme_updated.connect(self.set_pretty_display_theme)  # type: ignore

    def set_schema(self, schema: type[BaseModel]):
        self._md_schema = schema
        self.populate()

    def set_data(self, data: BaseModel):
        """Fill the data for the form.

        Args:
            data (BaseModel):   the data to enter into the form. Must be the same type as the
                                currently set schema, raises TypeError otherwise."""
        if not self._md_schema:
            raise ValueError("Schema not set - can't set data")
        if not isinstance(data, self._md_schema):
            raise TypeError(f"Supplied data {data} not of type {self._md_schema}")
        for form_item in self.enumerate_form_widgets():
            form_item.widget.setValue(getattr(data, form_item.label.property("_model_field_name")))

    def _form_item_specs(self):
        return [
            FormItemSpec(
                name=name, info=info, item_type=info.annotation, pretty_display=self._pretty_display
            )
            for name, info in self._md_schema.model_fields.items()
        ]

    def update_items_from_schema(self):
        self._items = self._form_item_specs()

    def populate(self):
        self.update_items_from_schema()
        super().populate()

    def get_form_data(self):
        """Get the entered metadata as a dict."""
        return self._dict_from_grid()

    def validate_form(self, *_) -> bool:
        """validate the currently entered metadata against the pydantic schema.
        If successful, returns on metadata_emitted and returns true.
        Otherwise, emits on form_data_cleared and returns false."""
        try:
            metadata_dict = self.get_form_data()
            self._md_schema.model_validate(metadata_dict)
            self._validity.set_global_state("success")
            self._validity_message.setText("No errors!")
            self.form_data_updated.emit(metadata_dict)
            self.validity_proc.emit(True)
            return True
        except ValidationError as e:
            self._validity.set_global_state("emergency")
            self._validity_message.setText(str(e))
            self.form_data_cleared.emit(None)
            self.validity_proc.emit(False)
            return False


class PydanticModelFormItem(DynamicFormItem):
    def __init__(
        self, parent: QWidget | None = None, *, spec: FormItemSpec, model: type[BaseModel]
    ) -> None:
        self._data_model = model

        super().__init__(parent=parent, spec=spec)
        self._main_widget.form_data_updated.connect(self._value_changed)

    def _add_main_widget(self) -> None:

        self._main_widget = PydanticModelForm(data_model=self._data_model)
        self._layout.addWidget(self._main_widget)

    def getValue(self):
        return self._main_widget.get_form_data()

    def setValue(self, value: dict):
        self._main_widget.set_data(self._data_model.model_validate(value))
